import numpy as np

import matplotlib.pyplot as pl
from matplotlib import rcParams
from matplotlib.colors import is_color_like

from scvelo.tools.utils import groups_to_bool
from scvelo.tools.velocity_embedding import (
    velocity_embedding as compute_velocity_embedding,
)
from .docs import doc_params, doc_scatter
from .scatter import scatter
from .utils import (
    default_arrow,
    default_basis,
    default_color,
    default_color_map,
    default_size,
    get_ax,
    get_components,
    get_figure_params,
    interpret_colorkey,
    make_unique_list,
    make_unique_valid_list,
    savefig_or_show,
    velocity_embedding_changed,
)


@doc_params(scatter=doc_scatter)
def velocity_embedding(
    adata,
    basis=None,
    vkey="velocity",
    density=None,
    arrow_size=None,
    arrow_length=None,
    scale=None,
    X=None,
    V=None,
    recompute=None,
    color=None,
    use_raw=None,
    layer=None,
    color_map=None,
    colorbar=True,
    palette=None,
    size=None,
    alpha=0.2,
    perc=None,
    sort_order=True,
    groups=None,
    components=None,
    projection="2d",
    legend_loc="none",
    legend_fontsize=None,
    legend_fontweight=None,
    xlabel=None,
    ylabel=None,
    title=None,
    fontsize=None,
    figsize=None,
    dpi=None,
    frameon=None,
    show=None,
    save=None,
    ax=None,
    ncols=None,
    **kwargs,
):
    """\
    Scatter plot of velocities on the embedding.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    density: `float` (default: 1)
        Amount of velocities to show - 0 none to 1 all
    arrow_size: `float` or triple `headlength, headwidth, headaxislength` (default: 1)
        Size of arrows.
    arrow_length: `float` (default: 1)
        Length of arrows.
    scale: `float` (default: 1)
        Length of velocities in the embedding.
    {scatter}

    Returns
    -------
    `matplotlib.Axis` if `show==False`
    """

    if vkey == "all":
        lkeys = list(adata.layers.keys())
        vkey = [key for key in lkeys if "velocity" in key and "_u" not in key]
    color, color_map = kwargs.pop("c", color), kwargs.pop("cmap", color_map)
    layers, vkeys = make_unique_list(layer), make_unique_list(vkey)
    colors = make_unique_list(color, allow_array=True)
    bases = make_unique_valid_list(adata, basis)
    bases = [default_basis(adata, **kwargs) if b is None else b for b in bases]

    if V is None:
        for key in vkeys:
            for bas in bases:
                if recompute or velocity_embedding_changed(adata, basis=bas, vkey=key):
                    compute_velocity_embedding(adata, basis=bas, vkey=key)

    scatter_kwargs = {
        "perc": perc,
        "use_raw": use_raw,
        "sort_order": sort_order,
        "alpha": alpha,
        "components": components,
        "projection": projection,
        "legend_loc": legend_loc,
        "groups": groups,
        "legend_fontsize": legend_fontsize,
        "legend_fontweight": legend_fontweight,
        "palette": palette,
        "color_map": color_map,
        "frameon": frameon,
        "xlabel": xlabel,
        "ylabel": ylabel,
        "colorbar": colorbar,
        "dpi": dpi,
        "fontsize": fontsize,
        "show": False,
        "save": False,
    }

    multikey = (
        colors
        if len(colors) > 1
        else layers
        if len(layers) > 1
        else vkeys
        if len(vkeys) > 1
        else bases
        if len(bases) > 1
        else None
    )
    if multikey is not None:
        if title is None:
            title = list(multikey)
        elif isinstance(title, (list, tuple)):
            title *= int(np.ceil(len(multikey) / len(title)))
        ncols = len(multikey) if ncols is None else min(len(multikey), ncols)
        nrows = int(np.ceil(len(multikey) / ncols))
        figsize = rcParams["figure.figsize"] if figsize is None else figsize
        figsize, dpi = get_figure_params(figsize, dpi, ncols)
        gs_figsize = (figsize[0] * ncols, figsize[1] * nrows)
        ax = []
        for i, gs in enumerate(
            pl.GridSpec(nrows, ncols, pl.figure(None, gs_figsize, dpi=dpi))
        ):
            if i < len(multikey):
                ax.append(
                    velocity_embedding(
                        adata,
                        density=density,
                        scale=scale,
                        size=size,
                        ax=pl.subplot(gs),
                        arrow_size=arrow_size,
                        arrow_length=arrow_length,
                        basis=bases[i] if len(bases) > 1 else basis,
                        color=colors[i] if len(colors) > 1 else color,
                        layer=layers[i] if len(layers) > 1 else layer,
                        vkey=vkeys[i] if len(vkeys) > 1 else vkey,
                        title=title[i] if isinstance(title, (list, tuple)) else title,
                        **scatter_kwargs,
                        **kwargs,
                    )
                )
        savefig_or_show(dpi=dpi, save=save, show=show)
        if show is False:
            return ax

    else:
        ax, show = get_ax(ax, show, figsize, dpi, projection)

        color, layer, vkey, basis = colors[0], layers[0], vkeys[0], bases[0]
        color = default_color(adata) if color is None else color
        color_map = default_color_map(adata, color) if color_map is None else color_map
        size = default_size(adata) / 2 if size is None else size
        if use_raw is None and "Ms" not in adata.layers.keys():
            use_raw = True
        _adata = (
            adata[groups_to_bool(adata, groups, groupby=color)]
            if groups is not None and color in adata.obs.keys()
            else adata
        )

        quiver_kwargs = {
            "scale": scale,
            "cmap": color_map,
            "angles": "xy",
            "scale_units": "xy",
            "edgecolors": "k",
            "linewidth": 0.1,
            "width": None,
        }
        if basis in adata.var_names:
            if use_raw:
                x = adata[:, basis].layers["spliced"]
                y = adata[:, basis].layers["unspliced"]
            else:
                x = adata[:, basis].layers["Ms"]
                y = adata[:, basis].layers["Mu"]
            dx = adata[:, basis].layers[vkey]
            dy = np.zeros(adata.n_obs)
            if f"{vkey}_u" in adata.layers.keys():
                dy = adata[:, basis].layers[f"{vkey}_u"]
            X = np.stack([np.ravel(x), np.ravel(y)]).T
            V = np.stack([np.ravel(dx), np.ravel(dy)]).T
        else:
            x = None if X is None else X[:, 0]
            y = None if X is None else X[:, 1]
            comps = get_components(components, basis, projection)
            X = _adata.obsm[f"X_{basis}"][:, comps] if X is None else X[:, :2]
            V = _adata.obsm[f"{vkey}_{basis}"][:, comps] if V is None else V[:, :2]

            hl, hw, hal = default_arrow(arrow_size)
            if arrow_length is not None:
                scale = 1 / arrow_length
            if scale is None:
                scale = 1
            quiver_kwargs.update({"scale": scale, "width": 0.0005, "headlength": hl})
            quiver_kwargs.update({"headwidth": hw, "headaxislength": hal})

        for arg in list(kwargs):
            if arg in quiver_kwargs:
                quiver_kwargs.update({arg: kwargs[arg]})
            else:
                scatter_kwargs.update({arg: kwargs[arg]})

        if (
            basis in adata.var_names
            and isinstance(color, str)
            and color in adata.layers.keys()
        ):
            c = interpret_colorkey(_adata, basis, color, perc)
        else:
            c = interpret_colorkey(_adata, color, layer, perc)

        if density is not None and 0 < density < 1:
            s = int(density * _adata.n_obs)
            ix_choice = np.random.choice(_adata.n_obs, size=s, replace=False)
            c = c[ix_choice] if len(c) == _adata.n_obs else c
            X = X[ix_choice]
            V = V[ix_choice]

        if projection == "3d" and X.shape[1] > 2 and V.shape[1] > 2:
            V, size = V / scale / 5, size / 10
            x0, x1, x2 = X[:, 0], X[:, 1], X[:, 2]
            v0, v1, v2 = V[:, 0], V[:, 1], V[:, 2]
            quiver3d_kwargs = {"zorder": 3, "linewidth": 0.5, "arrow_length_ratio": 0.3}
            c = list(c) + [element for element in list(c) for _ in range(2)]
            if is_color_like(c[0]):
                ax.quiver(x0, x1, x2, v0, v1, v2, color=c, **quiver3d_kwargs)
            else:
                ax.quiver(x0, x1, x2, v0, v1, v2, c, **quiver3d_kwargs)
        else:
            quiver_kwargs.update({"zorder": 3})
            if is_color_like(c[0]):
                ax.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], color=c, **quiver_kwargs)
            else:
                ax.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], c, **quiver_kwargs)

        scatter_kwargs.update({"basis": basis, "x": x, "y": y, "color": color})
        scatter_kwargs.update({"vkey": vkey, "layer": layer})
        ax = scatter(adata, size=size, title=title, ax=ax, zorder=0, **scatter_kwargs)

        savefig_or_show(dpi=dpi, save=save, show=show)
        if show is False:
            return ax



import inspect
import logging
import os
import sys
from datetime import datetime
from pathlib import Path, PurePosixPath
from typing import Dict, List, Mapping, Optional, Tuple, Union
from urllib.request import urlretrieve

import sphinx_autodoc_typehints
from docutils import nodes
from jinja2.defaults import DEFAULT_FILTERS
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.domains.python import PyObject, PyTypedField
from sphinx.environment import BuildEnvironment
from sphinx.ext import autosummary

import matplotlib  # noqa

HERE = Path(__file__).parent
sys.path.insert(0, str(HERE.parent.parent))
sys.path.insert(0, os.path.abspath("_ext"))

import scvelo  # isort:skip

# remove PyCharm’s old six module

if "six" in sys.modules:
    print(*sys.path, sep="\n")
    for pypath in list(sys.path):
        if any(p in pypath for p in ["PyCharm", "pycharm"]) and "helpers" in pypath:
            sys.path.remove(pypath)
    del sys.modules["six"]

matplotlib.use("agg")

logger = logging.getLogger(__name__)


# -- Basic notebooks and those stored under /vignettes and /perspectives --

notebooks_url = "https://github.com/theislab/scvelo_notebooks/raw/master/"
notebooks = []
notebook = [
    "VelocityBasics.ipynb",
    "DynamicalModeling.ipynb",
    "DifferentialKinetics.ipynb",
]
notebooks.extend(notebook)

notebook = [
    "Pancreas.ipynb",
    "DentateGyrus.ipynb",
    "NatureBiotechCover.ipynb",
    "Fig1_concept.ipynb",
    "Fig2_dentategyrus.ipynb",
    "Fig3_pancreas.ipynb",
    "FigS9_runtime.ipynb",
    "FigSuppl.ipynb",
]
notebooks.extend([f"vignettes/{nb}" for nb in notebook])

notebook = ["Perspectives.ipynb", "Perspectives_parameters.ipynb"]
notebooks.extend([f"perspectives/{nb}" for nb in notebook])

# -- Retrieve all notebooks --

for nb in notebooks:
    url = notebooks_url + nb
    try:
        urlretrieve(url, nb)
    except Exception as e:
        logger.error(f"Unable to retrieve notebook: `{url}`. Reason: `{e}`")
        pass


# -- General configuration ------------------------------------------------

needs_sphinx = "1.7"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx_autodoc_typehints",
    "nbsphinx",
    "edit_on_github",
]


# Generate the API documentation when building
autosummary_generate = True
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = False
napoleon_custom_sections = [("Params", "Parameters")]

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    anndata=("https://anndata.readthedocs.io/en/latest/", None),
    scanpy=("https://scanpy.readthedocs.io/en/latest/", None),
    cellrank=("https://cellrank.readthedocs.io/en/latest/", None),
)

templates_path = ["_templates"]
source_suffix = [".rst", ".ipynb"]
master_doc = "index"

# General information about the project.
project = "scVelo"
author = "Volker Bergen"
title = "RNA Velocity generalized through dynamical modeling"
copyright = f"{datetime.now():%Y}, {author}"

version = scvelo.__version__.replace(".dirty", "")
release = version
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
todo_include_todos = False

# Add notebooks prolog to Google Colab and nbviewer
nbsphinx_prolog = r"""
{% set docname = 'github/theislab/scvelo_notebooks/blob/master/' + env.doc2path(env.docname, base=None) %}
.. raw:: html

    <div class="note">
      <a href="https://colab.research.google.com/{{ docname|e }}" target="_parent">
      <img src="https://colab.research.google.com/assets/colab-badge.svg" alt="Open In Colab"/></a>
      <a href="https://nbviewer.jupyter.org/{{ docname|e }}" target="_parent">
      <img src="https://github.com/theislab/scvelo/raw/master/docs/source/_static/nbviewer-badge.svg" alt="Open In nbviewer"/></a>
    </div>
"""

# -- Options for HTML output ----------------------------------------------

html_theme = "sphinx_rtd_theme"
html_theme_options = dict(navigation_depth=1, titles_only=True)
github_repo = "scvelo"
github_nb_repo = "scvelo_notebooks"
html_static_path = ["_static"]


def setup(app):
    app.add_stylesheet("custom.css")


# -- Options for other output ------------------------------------------

htmlhelp_basename = "scvelodoc"
title_doc = f"{project} documentation"

latex_documents = [(master_doc, f"{project}.tex", title_doc, author, "manual")]
man_pages = [(master_doc, project, title_doc, [author], 1)]
texinfo_documents = [
    (master_doc, project, title_doc, author, project, title, "Miscellaneous")
]


# -- generate_options override ------------------------------------------


def process_generate_options(app: Sphinx):
    genfiles = app.config.autosummary_generate

    if genfiles and not hasattr(genfiles, "__len__"):
        env = app.builder.env
        genfiles = [
            env.doc2path(x, base=None)
            for x in env.found_docs
            if Path(env.doc2path(x)).is_file()
        ]
    if not genfiles:
        return

    from sphinx.ext.autosummary.generate import generate_autosummary_docs

    ext = app.config.source_suffix
    genfiles = [
        genfile + (not genfile.endswith(tuple(ext)) and ext[0] or "")
        for genfile in genfiles
    ]

    suffix = autosummary.get_rst_suffix(app)
    if suffix is None:
        return

    generate_autosummary_docs(
        genfiles,
        builder=app.builder,
        warn=logger.warning,
        info=logger.info,
        suffix=suffix,
        base_path=app.srcdir,
        imported_members=True,
        app=app,
    )


autosummary.process_generate_options = process_generate_options


# -- GitHub URLs for class and method pages ------------------------------------------


def get_obj_module(qualname):
    """Get a module/class/attribute and its original module by qualname"""
    modname = qualname
    classname = None
    attrname = None
    while modname not in sys.modules:
        attrname = classname
        modname, classname = modname.rsplit(".", 1)

    # retrieve object and find original module name
    if classname:
        cls = getattr(sys.modules[modname], classname)
        modname = cls.__module__
        obj = getattr(cls, attrname) if attrname else cls
    else:
        obj = None

    return obj, sys.modules[modname]


def get_linenos(obj):
    """Get an object’s line numbers"""
    try:
        lines, start = inspect.getsourcelines(obj)
    except TypeError:
        return None, None
    else:
        return start, start + len(lines) - 1


# set project_dir: project/docs/source/conf.py/../../.. → project/
project_dir = Path(__file__).parent.parent.parent
github_url_scvelo = "https://github.com/theislab/scvelo/tree/master"
github_url_read_loom = "https://github.com/theislab/anndata/tree/master/anndata"
github_url_read = "https://github.com/theislab/scanpy/tree/master"
github_url_scanpy = "https://github.com/theislab/scanpy/tree/master/scanpy"


def modurl(qualname):
    """Get the full GitHub URL for some object’s qualname."""
    obj, module = get_obj_module(qualname)
    github_url = github_url_scvelo
    try:
        path = PurePosixPath(Path(module.__file__).resolve().relative_to(project_dir))
    except ValueError:
        # trying to document something from another package
        github_url = (
            github_url_read_loom
            if "read_loom" in qualname
            else github_url_read
            if "read" in qualname
            else github_url_scanpy
        )
        path = "/".join(module.__file__.split("/")[-2:])
    start, end = get_linenos(obj)
    fragment = f"#L{start}-L{end}" if start and end else ""
    return f"{github_url}/{path}{fragment}"


def api_image(qualname: str) -> Optional[str]:
    path = Path(__file__).parent / f"{qualname}.png"
    print(path, path.is_file())
    return (
        f".. image:: {path.name}\n   :width: 200\n   :align: right"
        if path.is_file()
        else ""
    )


# modify the default filters

DEFAULT_FILTERS.update(modurl=modurl, api_image=api_image)

# -- Override some classnames in autodoc --------------------------------------------

qualname_overrides = {
    "anndata.base.AnnData": "anndata.AnnData",
    "scvelo.pl.scatter": "scvelo.plotting.scatter",
}

fa_orig = sphinx_autodoc_typehints.format_annotation


def format_annotation(annotation):
    if getattr(annotation, "__origin__", None) is Union or hasattr(
        annotation, "__union_params__"
    ):
        params = getattr(annotation, "__union_params__", None) or getattr(
            annotation, "__args__", None
        )
        return ", ".join(map(format_annotation, params))
    if getattr(annotation, "__origin__", None) is Mapping:
        return ":class:`~typing.Mapping`"
    if inspect.isclass(annotation):
        full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        override = qualname_overrides.get(full_name)
        if override is not None:
            return f":py:class:`~{qualname_overrides[full_name]}`"
    return fa_orig(annotation)


sphinx_autodoc_typehints.format_annotation = format_annotation


# -- Prettier Param docs --------------------------------------------


class PrettyTypedField(PyTypedField):
    list_type = nodes.definition_list

    def make_field(
        self,
        types: Dict[str, List[nodes.Node]],
        domain: str,
        items: Tuple[str, List[nodes.inline]],
        env: BuildEnvironment = None,
    ) -> nodes.field:
        def makerefs(rolename, name, node):
            return self.make_xrefs(rolename, domain, name, node, env=env)

        def handle_item(
            fieldarg: str, content: List[nodes.inline]
        ) -> nodes.definition_list_item:
            head = nodes.term()
            head += makerefs(self.rolename, fieldarg, addnodes.literal_strong)
            fieldtype = types.pop(fieldarg, None)
            if fieldtype is not None:
                head += nodes.Text(" : ")
                if len(fieldtype) == 1 and isinstance(fieldtype[0], nodes.Text):
                    (text_node,) = fieldtype  # type: nodes.Text
                    head += makerefs(
                        self.typerolename, text_node.astext(), addnodes.literal_emphasis
                    )
                else:
                    head += fieldtype

            body_content = nodes.paragraph("", "", *content)
            body = nodes.definition("", body_content)

            return nodes.definition_list_item("", head, body)

        fieldname = nodes.field_name("", self.label)
        if len(items) == 1 and self.can_collapse:
            fieldarg, content = items[0]
            bodynode = handle_item(fieldarg, content)
        else:
            bodynode = self.list_type()
            for fieldarg, content in items:
                bodynode += handle_item(fieldarg, content)
        fieldbody = nodes.field_body("", bodynode)
        return nodes.field("", fieldname, fieldbody)


# replace matching field types with ours
PyObject.doc_field_types = [
    PrettyTypedField(
        ft.name,
        names=ft.names,
        typenames=ft.typenames,
        label=ft.label,
        rolename=ft.rolename,
        typerolename=ft.typerolename,
        can_collapse=ft.can_collapse,
    )
    if isinstance(ft, PyTypedField)
    else ft
    for ft in PyObject.doc_field_types
]



from typing import Dict, List, Tuple, Union

import numpy as np
from numpy import ndarray

from ._arithmetic import invert
from ._base import DynamicsBase


# TODO: Improve parameter names: alpha -> transcription_rate; beta -> splicing_rate;
# gamma -> degradation_rate
# TODO: Handle cases beta = 0, gamma == 0, beta == gamma
class SplicingDynamics(DynamicsBase):
    """Splicing dynamics.

    Arguments
    ---------
    alpha
        Transcription rate.
    beta
        Translation rate.
    gamma
        Splicing degradation rate.
    initial_state
        Initial state of system. Defaults to `[0, 0]`.

    Attributes
    ----------
    alpha
        Transcription rate.
    beta
        Translation rate.
    gamma
        Splicing degradation rate.
    initial_state
        Initial state of system. Defaults to `[0, 0]`.
    u0
        Initial abundance of unspliced RNA.
    s0
        Initial abundance of spliced RNA.

    """

    def __init__(
        self,
        alpha: float,
        beta: float,
        gamma: float,
        initial_state: Union[List, ndarray] = [0, 0],
    ):

        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

        self.initial_state = initial_state

    @property
    def initial_state(self):
        return self._initial_state

    @initial_state.setter
    def initial_state(self, val):
        if isinstance(val, list) or (isinstance(val, ndarray) and (val.ndim == 1)):
            self.u0 = val[0]
            self.s0 = val[1]
        else:
            self.u0 = val[:, 0]
            self.s0 = val[:, 1]
        self._initial_state = val

    def get_solution(
        self, t: ndarray, stacked: bool = True, with_keys: bool = False
    ) -> Union[Dict, ndarray]:
        """Calculate solution of dynamics.

        Arguments
        ---------
        t
            Time steps at which to evaluate solution.
        stacked
            Whether to stack states or return them individually. Defaults to `True`.
        with_keys
            Whether to return solution labelled by variables in form of a dictionary.
            Defaults to `False`.

        Returns
        -------
        Union[Dict, ndarray]
            Solution of system. If `with_keys=True`, the solution is returned in form of
            a dictionary with variables as keys. Otherwise, the solution is given as
            a `numpy.ndarray` of form `(n_steps, 2)`.
        """

        expu = np.exp(-self.beta * t)
        exps = np.exp(-self.gamma * t)

        unspliced = self.u0 * expu + self.alpha / self.beta * (1 - expu)
        c = (self.alpha - self.u0 * self.beta) * invert(self.gamma - self.beta)
        spliced = (
            self.s0 * exps + self.alpha / self.gamma * (1 - exps) + c * (exps - expu)
        )

        if with_keys:
            return {"u": unspliced, "s": spliced}
        elif not stacked:
            return unspliced, spliced
        else:
            if isinstance(t, np.ndarray) and t.ndim == 2:
                return np.stack([unspliced, spliced], axis=2)
            else:
                return np.column_stack([unspliced, spliced])

    # TODO: Handle cases `beta = 0`, `gamma = 0`
    def get_steady_states(
        self, stacked: bool = True, with_keys: bool = False
    ) -> Union[Dict[str, ndarray], Tuple[ndarray], ndarray]:
        """Return steady state of system.

        Arguments
        ---------
        stacked
            Whether to stack states or return them individually. Defaults to `True`.
        with_keys
            Whether to return solution labelled by variables in form of a dictionary.
            Defaults to `False`.

        Returns
        -------
        Union[Dict[str, ndarray], Tuple[ndarray], ndarray]
            Steady state of system.
        """

        if (self.beta > 0) and (self.gamma > 0):
            unspliced = self.alpha / self.beta
            spliced = self.alpha / self.gamma

        if with_keys:
            return {"u": unspliced, "s": spliced}
        elif not stacked:
            return unspliced, spliced
        else:
            return np.array([unspliced, spliced])



import re
from typing import List, Optional, Union

from typing_extensions import Literal

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from pandas.api.types import is_categorical_dtype
from scipy.sparse import csr_matrix, issparse, spmatrix

from anndata import AnnData

from scvelo import logging as logg
from ._arithmetic import sum


def clean_obs_names(
    data: AnnData,
    base: str = "[AGTCBDHKMNRSVWY]",
    ID_length: int = 12,
    copy: bool = False,
) -> Optional[AnnData]:
    """Clean up the obs_names.

    For example an obs_name 'sample1_AGTCdate' is changed to 'AGTC' of the sample
    'sample1_date'. The sample name is then saved in obs['sample_batch'].
    The genetic codes are identified according to according to
    https://www.neb.com/tools-and-resources/usage-guidelines/the-genetic-code.

    Arguments
    ---------
    data
        Annotated data matrix.
    base
        Genetic code letters to be identified.
    ID_length
        Length of the Genetic Codes in the samples.
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    Optional[AnnData]
        Returns or updates `adata` with the attributes
        obs_names: list
            updated names of the observations
        sample_batch: `.obs`
            names of the identified sample batches
    """

    def get_base_list(name, base):
        base_list = base
        while re.search(base_list + base, name) is not None:
            base_list += base
        if len(base_list) == 0:
            raise ValueError("Encountered an invalid ID in obs_names: ", name)
        return base_list

    adata = data.copy() if copy else data

    names = adata.obs_names
    base_list = get_base_list(names[0], base)

    if len(np.unique([len(name) for name in adata.obs_names])) == 1:
        start, end = re.search(base_list, names[0]).span()
        newIDs = [name[start:end] for name in names]
        start, end = 0, len(newIDs[0])
        for i in range(end - ID_length):
            if np.any([ID[i] not in base for ID in newIDs]):
                start += 1
            if np.any([ID[::-1][i] not in base for ID in newIDs]):
                end -= 1

        newIDs = [ID[start:end] for ID in newIDs]
        prefixes = [names[i].replace(newIDs[i], "") for i in range(len(names))]
    else:
        prefixes, newIDs = [], []
        for name in names:
            match = re.search(base_list, name)
            newID = (
                re.search(get_base_list(name, base), name).group()
                if match is None
                else match.group()
            )
            newIDs.append(newID)
            prefixes.append(name.replace(newID, ""))

    adata.obs_names = newIDs
    if len(prefixes[0]) > 0 and len(np.unique(prefixes)) > 1:
        adata.obs["sample_batch"] = (
            pd.Categorical(prefixes)
            if len(np.unique(prefixes)) < adata.n_obs
            else prefixes
        )

    adata.obs_names_make_unique()
    return adata if copy else None


def cleanup(
    data: AnnData,
    clean: Union[
        Literal["layers", "obs", "var", "uns"],
        List[Literal["layers", "obs", "var", "uns"]],
    ] = "layers",
    keep: Optional[Union[str, List[str]]] = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """Delete not needed attributes.

    Arguments
    ---------
    data
        Annotated data matrix.
    clean
        Which attributes to consider for freeing memory.
    keep
        Which attributes to keep.
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    Optional[AnnData]
        Returns or updates `adata` with selection of attributes kept.
    """

    adata = data.copy() if copy else data
    verify_dtypes(adata)

    keep = list([keep] if isinstance(keep, str) else {} if keep is None else keep)
    keep.extend(["spliced", "unspliced", "Ms", "Mu", "clusters", "neighbors"])

    ann_dict = {
        "obs": adata.obs_keys(),
        "var": adata.var_keys(),
        "uns": adata.uns_keys(),
        "layers": list(adata.layers.keys()),
    }

    if "all" not in clean:
        ann_dict = {ann: values for (ann, values) in ann_dict.items() if ann in clean}

    for (ann, values) in ann_dict.items():
        for value in values:
            if value not in keep:
                del getattr(adata, ann)[value]

    return adata if copy else None


def get_df(
    data: AnnData,
    keys: Optional[Union[str, List[str]]] = None,
    layer: Optional[str] = None,
    index: List = None,
    columns: List = None,
    sort_values: bool = None,
    dropna: Literal["all", "any"] = "all",
    precision: int = None,
) -> DataFrame:
    """Get dataframe for a specified adata key.

    Return values for specified key
    (in obs, var, obsm, varm, obsp, varp, uns, or layers) as a dataframe.

    Arguments
    ---------
    data
        AnnData object or a numpy array to get values from.
    keys
        Keys from `.var_names`, `.obs_names`, `.var`, `.obs`,
        `.obsm`, `.varm`, `.obsp`, `.varp`, `.uns`, or `.layers`.
    layer
        Layer of `adata` to use as expression values.
    index
        List to set as index.
    columns
        List to set as columns names.
    sort_values
        Wether to sort values by first column (sort_values=True) or a specified column.
    dropna
        Drop columns/rows that contain NaNs in all ('all') or in any entry ('any').
    precision
        Set precision for pandas dataframe.

    Returns
    -------
    :class:`pd.DataFrame`
        A dataframe.
    """

    if precision is not None:
        pd.set_option("precision", precision)

    if isinstance(data, AnnData):
        keys, keys_split = (
            keys.split("*") if isinstance(keys, str) and "*" in keys else (keys, None)
        )
        keys, key_add = (
            keys.split("/") if isinstance(keys, str) and "/" in keys else (keys, None)
        )
        keys = [keys] if isinstance(keys, str) else keys
        key = keys[0]

        s_keys = ["obs", "var", "obsm", "varm", "uns", "layers"]
        d_keys = [
            data.obs.keys(),
            data.var.keys(),
            data.obsm.keys(),
            data.varm.keys(),
            data.uns.keys(),
            data.layers.keys(),
        ]

        if hasattr(data, "obsp") and hasattr(data, "varp"):
            s_keys.extend(["obsp", "varp"])
            d_keys.extend([data.obsp.keys(), data.varp.keys()])

        if keys is None:
            df = data.to_df()
        elif key in data.var_names:
            df = obs_df(data, keys, layer=layer)
        elif key in data.obs_names:
            df = var_df(data, keys, layer=layer)
        else:
            if keys_split is not None:
                keys = [
                    k
                    for k in list(data.obs.keys()) + list(data.var.keys())
                    if key in k and keys_split in k
                ]
                key = keys[0]
            s_key = [s for (s, d_key) in zip(s_keys, d_keys) if key in d_key]
            if len(s_key) == 0:
                raise ValueError(f"'{key}' not found in any of {', '.join(s_keys)}.")
            if len(s_key) > 1:
                logg.warn(f"'{key}' found multiple times in {', '.join(s_key)}.")

            s_key = s_key[-1]
            df = getattr(data, s_key)[keys if len(keys) > 1 else key]
            if key_add is not None:
                df = df[key_add]
            if index is None:
                index = (
                    data.var_names
                    if s_key == "varm"
                    else data.obs_names
                    if s_key in {"obsm", "layers"}
                    else None
                )
                if index is None and s_key == "uns" and hasattr(df, "shape"):
                    key_cats = np.array(
                        [
                            key
                            for key in data.obs.keys()
                            if is_categorical_dtype(data.obs[key])
                        ]
                    )
                    num_cats = [
                        len(data.obs[key].cat.categories) == df.shape[0]
                        for key in key_cats
                    ]
                    if np.sum(num_cats) == 1:
                        index = data.obs[key_cats[num_cats][0]].cat.categories
                        if (
                            columns is None
                            and len(df.shape) > 1
                            and df.shape[0] == df.shape[1]
                        ):
                            columns = index
            elif isinstance(index, str) and index in data.obs.keys():
                index = pd.Categorical(data.obs[index]).categories
            if columns is None and s_key == "layers":
                columns = data.var_names
            elif isinstance(columns, str) and columns in data.obs.keys():
                columns = pd.Categorical(data.obs[columns]).categories
    elif isinstance(data, pd.DataFrame):
        if isinstance(keys, str) and "*" in keys:
            keys, keys_split = keys.split("*")
            keys = [k for k in data.columns if keys in k and keys_split in k]
        df = data[keys] if keys is not None else data
    else:
        df = data

    if issparse(df):
        df = np.array(df.A)
    if columns is None and hasattr(df, "names"):
        columns = df.names

    df = pd.DataFrame(df, index=index, columns=columns)

    if dropna:
        df.replace("", np.nan, inplace=True)
        how = dropna if isinstance(dropna, str) else "any" if dropna is True else "all"
        df.dropna(how=how, axis=0, inplace=True)
        df.dropna(how=how, axis=1, inplace=True)

    if sort_values:
        sort_by = (
            sort_values
            if isinstance(sort_values, str) and sort_values in df.columns
            else df.columns[0]
        )
        df = df.sort_values(by=sort_by, ascending=False)

    if hasattr(data, "var_names"):
        if df.index[0] in data.var_names:
            df.var_names = df.index
        elif df.columns[0] in data.var_names:
            df.var_names = df.columns
    if hasattr(data, "obs_names"):
        if df.index[0] in data.obs_names:
            df.obs_names = df.index
        elif df.columns[0] in data.obs_names:
            df.obs_names = df.columns

    return df


# TODO: Generalize to arbitrary modality
def get_initial_size(
    adata: AnnData, layer: Optional[str] = None, by_total_size: bool = False
) -> Optional[ndarray]:
    """Get initial counts per observation of a layer.

    Arguments
    ---------
    adata
        Annotated data matrix.
    layer
        Name of layer for which to retrieve initial size.
    by_total_size
        Whether or not to return the combined initial size of the spliced and unspliced
        layers.

    Returns
    -------
    np.ndarray
        Initial counts per observation in the specified layer.
    """

    if by_total_size:
        sizes = [
            adata.obs[f"initial_size_{layer}"]
            for layer in {"spliced", "unspliced"}
            if f"initial_size_{layer}" in adata.obs.keys()
        ]
        return np.sum(sizes, axis=0)
    elif layer in adata.layers.keys():
        return (
            np.array(adata.obs[f"initial_size_{layer}"])
            if f"initial_size_{layer}" in adata.obs.keys()
            else get_size(adata, layer)
        )
    elif layer is None or layer == "X":
        return (
            np.array(adata.obs["initial_size"])
            if "initial_size" in adata.obs.keys()
            else get_size(adata)
        )
    else:
        return None


def get_modality(adata: AnnData, modality: str) -> Union[ndarray, spmatrix]:
    """Extract data of one modality.

    Arguments
    ---------
    adata
        Annotated data to extract modality from.
    modality
        Modality for which data is needed.

    Returns
    -------
    Union[ndarray, spmatrix]
        Retrieved modality from :class:`~anndata.AnnData` object.
    """

    if modality == "X":
        return adata.X
    elif modality in adata.layers.keys():
        return adata.layers[modality]
    elif modality in adata.obsm.keys():
        if isinstance(adata.obsm[modality], DataFrame):
            return adata.obsm[modality].values
        else:
            return adata.obsm[modality]


# TODO: Generalize to arbitray modality
def get_size(adata: AnnData, layer: Optional[str] = None) -> ndarray:
    """Get counts per observation in a layer.

    Arguments
    ---------
    adata
        Annotated data matrix.
    layer
        Name of later for which to retrieve initial size.

    Returns
    -------
    np.ndarray
        Initial counts per observation in the specified layer.
    """

    X = adata.X if layer is None else adata.layers[layer]
    return sum(X, axis=1)


def make_dense(
    adata: AnnData, modalities: Union[List[str], str], inplace: bool = True
) -> Optional[AnnData]:
    """Densify sparse AnnData entry.

    Arguments
    ---------
    adata
        Annotated data object.
    modality
        Modality to make dense.
    inplace
        Boolean flag to perform operations inplace or not. Defaults to `True`.

    Returns
    -------
    Optional[AnnData]
        Copy of annotated data `adata` if `inplace=True` with dense modalities.
    """

    if not inplace:
        adata = adata.copy()

    if isinstance(modalities, str):
        modalities = [modalities]

    # Densify modalities
    for modality in modalities:
        count_data = get_modality(adata=adata, modality=modality)
        if issparse(count_data):
            set_modality(adata=adata, modality=modality, new_value=count_data.A)

    return adata if not inplace else None


# TODO: Allow choosing format of sparse matrix i.e., csr, csc, ...
def make_sparse(
    adata: AnnData, modalities: Union[List[str], str], inplace: bool = True
) -> Optional[AnnData]:
    """Make AnnData entry sparse.

    Arguments
    ---------
    adata
        Annotated data object.
    modality
        Modality to make sparse.
    inplace
        Boolean flag to perform operations inplace or not. Defaults to `True`.

    Returns
    -------
    Optional[AnnData]
        Copy of annotated data `adata` with sparse modalities if `inplace=True`.
    """

    if not inplace:
        adata = adata.copy()

    if isinstance(modalities, str):
        modalities = [modalities]

    # Make modalities sparse
    for modality in modalities:
        count_data = get_modality(adata=adata, modality=modality)
        if modality == "X":
            logg.warn("Making `X` sparse is not supported.")
        elif not issparse(count_data):
            set_modality(
                adata=adata, modality=modality, new_value=csr_matrix(count_data)
            )

    return adata if not inplace else None


def merge(adata: AnnData, ldata: AnnData, copy: bool = True) -> Optional[AnnData]:
    """Merge two annotated data matrices.

    Arguments
    ---------
    adata
        Annotated data matrix (reference data set).
    ldata
        Annotated data matrix (to be merged into adata).
    copy
        Boolean flag to manipulate original AnnData or a copy of it.

    Returns
    -------
    Optional[:class:`anndata.AnnData`]
        Returns a :class:`~anndata.AnnData` object
    """

    adata.var_names_make_unique()
    ldata.var_names_make_unique()

    if (
        "spliced" in ldata.layers.keys()
        and "initial_size_spliced" not in ldata.obs.keys()
    ):
        set_initial_size(ldata)
    elif (
        "spliced" in adata.layers.keys()
        and "initial_size_spliced" not in adata.obs.keys()
    ):
        set_initial_size(adata)

    common_obs = pd.unique(adata.obs_names.intersection(ldata.obs_names))
    common_vars = pd.unique(adata.var_names.intersection(ldata.var_names))

    if len(common_obs) == 0:
        clean_obs_names(adata)
        clean_obs_names(ldata)
        common_obs = adata.obs_names.intersection(ldata.obs_names)

    if copy:
        _adata = adata[common_obs].copy()
        _ldata = ldata[common_obs].copy()
    else:
        adata._inplace_subset_obs(common_obs)
        _adata, _ldata = adata, ldata[common_obs].copy()

    _adata.var_names_make_unique()
    _ldata.var_names_make_unique()

    same_vars = len(_adata.var_names) == len(_ldata.var_names) and np.all(
        _adata.var_names == _ldata.var_names
    )
    join_vars = len(common_vars) > 0

    if join_vars and not same_vars:
        _adata._inplace_subset_var(common_vars)
        _ldata._inplace_subset_var(common_vars)

    for attr in _ldata.obs.keys():
        if attr not in _adata.obs.keys():
            _adata.obs[attr] = _ldata.obs[attr]
    for attr in _ldata.obsm.keys():
        if attr not in _adata.obsm.keys():
            _adata.obsm[attr] = _ldata.obsm[attr]
    for attr in _ldata.uns.keys():
        if attr not in _adata.uns.keys():
            _adata.uns[attr] = _ldata.uns[attr]
    if join_vars:
        for attr in _ldata.layers.keys():
            if attr not in _adata.layers.keys():
                _adata.layers[attr] = _ldata.layers[attr]

        if _adata.shape[1] == _ldata.shape[1]:
            same_vars = len(_adata.var_names) == len(_ldata.var_names) and np.all(
                _adata.var_names == _ldata.var_names
            )
            if same_vars:
                for attr in _ldata.var.keys():
                    if attr not in _adata.var.keys():
                        _adata.var[attr] = _ldata.var[attr]
                for attr in _ldata.varm.keys():
                    if attr not in _adata.varm.keys():
                        _adata.varm[attr] = _ldata.varm[attr]
            else:
                raise ValueError("Variable names are not identical.")

    return _adata if copy else None


def obs_df(adata: AnnData, keys: List[str], layer: Optional[str] = None) -> DataFrame:
    """Extract layer as Pandas DataFrame indexed by observation.

    Arguments
    ---------
    adata
        Annotated data matrix (reference data set).
    keys
        Variables for which to extract data.
    layer
        Name of layer to turn into a Pandas DataFrame.

    Returns
    -------
    DataFrame
        DataFrame indexed by observations. Columns correspond to variables of specified
        layer.
    """

    lookup_keys = [k for k in keys if k in adata.var_names]
    if len(lookup_keys) < len(keys):
        logg.warn(
            f"Keys {[k for k in keys if k not in adata.var_names]} "
            f"were not found in `adata.var_names`."
        )

    df = pd.DataFrame(index=adata.obs_names)
    for lookup_key in lookup_keys:
        df[lookup_key] = adata.obs_vector(lookup_key, layer=layer)
    return df


# TODO: Generalize to arbitrary modality
def set_initial_size(adata: AnnData, layers: Optional[str] = None) -> None:
    """Set current counts per observation of a layer as its initial size.

    The initial size is only set if it does not already exist.

    Arguments
    ---------
    adata
        Annotated data matrix.
    layers
        Name of layers for which to calculate initial size.

    Returns
    -------
    None
    """

    if layers is None:
        layers = ["spliced", "unspliced"]
    verify_dtypes(adata)
    layers = [
        layer
        for layer in layers
        if layer in adata.layers.keys()
        and f"initial_size_{layer}" not in adata.obs.keys()
    ]
    for layer in layers:
        adata.obs[f"initial_size_{layer}"] = get_size(adata, layer)
    if "initial_size" not in adata.obs.keys():
        adata.obs["initial_size"] = get_size(adata)


def set_modality(
    adata: AnnData,
    new_value: Union[ndarray, spmatrix, DataFrame],
    modality: Optional[str] = None,
    inplace: bool = True,
) -> Optional[AnnData]:
    """Set modality of annotated data object to new value.

    Arguments
    ---------
    adata
        Annotated data object.
    new_value
        New value of modality.
    modality
        Modality to overwrite with new value. Defaults to `None`.
    inplace
        Boolean flag to indicate whether setting of modality should be inplace or
            not. Defaults to `True`.

    Returns
    -------
    Optional[AnnData]
        Copy of annotated data `adata` with updated modality if `inplace=True`.
    """

    if not inplace:
        adata = adata.copy()

    if (modality == "X") or (modality is None):
        adata.X = new_value
    elif modality in adata.layers.keys():
        adata.layers[modality] = new_value
    elif modality in adata.obsm.keys():
        adata.obsm[modality] = new_value

    if not inplace:
        return adata


def show_proportions(
    adata: AnnData, layers: Optional[str] = None, use_raw: bool = True
) -> None:
    """Proportions of abundances of modalities in layers.

    The proportions are printed.

    Arguments
    ---------
    adata
        Annotated data matrix.
    layers
        Layers to consider.
    use_raw
        Use initial sizes, i.e., raw data, to determine proportions.

    Returns
    -------
    None
    """

    if layers is None:
        layers = ["spliced", "unspliced", "ambigious"]
    layers_keys = [key for key in layers if key in adata.layers.keys()]
    counts_layers = [sum(adata.layers[key], axis=1) for key in layers_keys]
    if use_raw:
        size_key, obs = "initial_size_", adata.obs
        counts_layers = [
            obs[size_key + layer] if size_key + layer in obs.keys() else c
            for layer, c in zip(layers_keys, counts_layers)
        ]

    counts_per_cell_sum = np.sum(counts_layers, 0)
    counts_per_cell_sum += counts_per_cell_sum == 0

    mean_abundances = [
        np.mean(counts_per_cell / counts_per_cell_sum)
        for counts_per_cell in counts_layers
    ]

    print(f"Abundance of {layers_keys}: {np.round(mean_abundances, 2)}")


def var_df(adata: AnnData, keys: List[str], layer: Optional[str] = None):
    """Extract layer as Pandas DataFrame indexed by features.

    Arguments
    ---------
    adata
        Annotated data matrix (reference data set).
    keys
        Observations for which to extract data.
    layer
        Name of layer to turn into a Pandas DataFrame.

    Returns
    -------
    DataFrame
        DataFrame indexed by features. Columns correspond to observations of specified
        layer.
    """

    lookup_keys = [k for k in keys if k in adata.obs_names]
    if len(lookup_keys) < len(keys):
        logg.warn(
            f"Keys {[k for k in keys if k not in adata.obs_names]} "
            f"were not found in `adata.obs_names`."
        )

    df = pd.DataFrame(index=adata.var_names)
    for lookup_key in lookup_keys:
        df[lookup_key] = adata.var_vector(lookup_key, layer=layer)
    return df


# TODO: Find better function name
def verify_dtypes(adata: AnnData) -> None:
    """Verify that AnnData object is not corrupted.

    Arguments
    ---------
    adata
        Annotated data matrix to check.

    Returns
    -------
    None
    """

    try:
        _ = adata[:, 0]
    except Exception:
        uns = adata.uns
        adata.uns = {}
        try:
            _ = adata[:, 0]
            logg.warn(
                "Safely deleted unstructured annotations (adata.uns), \n"
                "as these do not comply with permissible anndata datatypes."
            )
        except Exception:
            logg.warn(
                "The data might be corrupted. Please verify all annotation datatypes."
            )
            adata.uns = uns



from .moments import moments
from .neighbors import neighbors, pca, remove_duplicate_cells
from .utils import (
    cleanup,
    filter_and_normalize,
    filter_genes,
    filter_genes_dispersion,
    log1p,
    normalize_per_cell,
    recipe_velocity,
    show_proportions,
)

__all__ = [
    "cleanup",
    "filter_and_normalize",
    "filter_genes",
    "filter_genes_dispersion",
    "log1p",
    "moments",
    "neighbors",
    "normalize_per_cell",
    "pca",
    "recipe_velocity",
    "remove_duplicate_cells",
    "show_proportions",
]



"""scvelo - RNA velocity generalized through dynamical modeling"""
from anndata import AnnData
from scanpy import read, read_loom

from scvelo import datasets, logging, pl, pp, settings, tl, utils
from scvelo.core import get_df
from scvelo.plotting.gridspec import GridSpec
from scvelo.preprocessing.neighbors import Neighbors
from scvelo.read_load import DataFrame, load, read_csv
from scvelo.settings import set_figure_params
from scvelo.tools.run import run_all, test
from scvelo.tools.utils import round
from scvelo.tools.velocity import Velocity
from scvelo.tools.velocity_graph import VelocityGraph

try:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)
    del get_version
except (LookupError, ImportError):
    try:
        from importlib_metadata import version  # Python < 3.8
    except Exception:
        from importlib.metadata import version  # Python = 3.8
    __version__ = version(__name__)
    del version


__all__ = [
    "AnnData",
    "DataFrame",
    "datasets",
    "get_df",
    "GridSpec",
    "load",
    "logging",
    "Neighbors",
    "pl",
    "pp",
    "read",
    "read_csv",
    "read_loom",
    "round",
    "run_all",
    "set_figure_params",
    "settings",
    "test",
    "tl",
    "utils",
    "Velocity",
    "VelocityGraph",
]



import numpy as np
import pandas as pd

from scvelo.tools.dynamical_model import latent_time
from scvelo.tools.rank_velocity_genes import rank_velocity_genes
from scvelo.tools.score_genes_cell_cycle import score_genes_cell_cycle
from scvelo.tools.velocity_pseudotime import velocity_pseudotime
from .gridspec import GridSpec
from .utils import make_unique_list


def summary(adata, basis="umap", color="clusters", n_top_genes=12, var_names=None):
    if "fit_alpha" in adata.var.keys():
        tkey = "latent_time"
        if tkey not in adata.obs.keys():
            latent_time(adata)
        top_genes = adata.var.sort_values("fit_likelihood", ascending=False)
        top_genes = top_genes.index[:n_top_genes]
    else:
        tkey = "velocity_pseudotime"
        velocity_pseudotime(adata)
        if "rank_velocity_genes" not in adata.uns.keys():
            rank_velocity_genes(adata, groupby=color)
        top_genes = pd.DataFrame(adata.uns["rank_velocity_genes"]["names"])
        top_genes = top_genes.iloc[0][:n_top_genes]
    nrows = 1 + int(np.ceil(len(top_genes) / 4))

    if "S_score" not in adata.obs.keys() or "G2M_score" not in adata.obs.keys():
        score_genes_cell_cycle(adata)
    adata.obs["S"] = np.clip(adata.obs["S_score"], 0.1, None)
    adata.obs["G2M"] = np.clip(adata.obs["G2M_score"], 0.1, None)

    if var_names is not None:
        var_names = make_unique_list(var_names)
        var_names = [name for name in var_names if name in adata.var_names]
        nrows += int(np.ceil(len(var_names) / 4))

    kwargs = dict(
        c=color, legend_loc_lines="none", add_outline="fit_diff_kinetics", frameon=False
    )
    with GridSpec(ncols=4, nrows=nrows) as pl:
        pl.scatter(adata, basis=basis, c=color)
        pl.velocity_embedding_stream(adata, c=tkey)
        pl.scatter(adata, color_gradients=["root_cells", "end_points"], palette="jet")
        pl.scatter(adata, color_gradients=["S", "G2M"])

        if top_genes is not None:
            for gene in top_genes:
                text = r"$L=$ " + f"{np.round(adata[:, gene].var.fit_likelihood[0], 2)}"
                pl.scatter(adata, basis=gene, add_text=text, **kwargs)

        if var_names is not None:
            for gene in var_names:
                text = r"$L=$ " + f"{np.round(adata[:, gene].var.fit_likelihood[0], 2)}"
                pl.scatter(adata, basis=gene, add_text=text, **kwargs)



import numpy as np
from scipy.sparse import issparse

from scvelo import logging as logg
from scvelo import settings
from .utils import strings_to_categoricals, vcorrcoef
from .velocity_pseudotime import velocity_pseudotime


def get_mean_var(X, ignore_zeros=False, perc=None):
    data = X.data if issparse(X) else X
    mask_nans = np.isnan(data) | np.isinf(data) | np.isneginf(data)

    n_nonzeros = (X != 0).sum(0)
    n_counts = n_nonzeros if ignore_zeros else X.shape[0]

    if mask_nans.sum() > 0:
        if issparse(X):
            data[np.isnan(data) | np.isinf(data) | np.isneginf(data)] = 0
            n_nans = n_nonzeros - (X != 0).sum(0)
        else:
            X[mask_nans] = 0
            n_nans = mask_nans.sum(0)
        n_counts -= n_nans

    if perc is not None:
        if np.size(perc) < 2:
            perc = [perc, 100] if perc < 50 else [0, perc]
        lb, ub = np.percentile(data, perc)
        data = np.clip(data, lb, ub)

    if issparse(X):
        mean = (X.sum(0) / n_counts).A1
        mean_sq = (X.multiply(X).sum(0) / n_counts).A1
    else:
        mean = X.sum(0) / n_counts
        mean_sq = np.multiply(X, X).sum(0) / n_counts
    n_cells = np.clip(X.shape[0], 2, None)  # to avoid division by zero
    var = (mean_sq - mean ** 2) * (n_cells / (n_cells - 1))

    mean = np.nan_to_num(mean)
    var = np.nan_to_num(var)
    return mean, var


def select_groups(adata, groups="all", key="louvain"):
    """Get subset of groups in adata.obs[key]."""
    strings_to_categoricals(adata)
    if isinstance(groups, list) and isinstance(groups[0], int):
        groups = [f"{n}" for n in groups]
    categories = adata.obs[key].cat.categories
    groups_masks = np.array(
        [categories[i] == adata.obs[key].values for i, name in enumerate(categories)]
    )
    if groups == "all":
        groups = categories.values
    else:
        groups_ids = [categories.get_loc(name) for name in groups]
        groups_masks = groups_masks[groups_ids]
        groups = categories[groups_ids].values
    return groups, groups_masks


def velocity_clusters(
    data,
    vkey="velocity",
    match_with="clusters",
    sort_by="velocity_pseudotime",
    resolution=None,
    min_likelihood=None,
    copy=False,
):
    """Computes velocity clusters via louvain on velocities.

    .. code:: python

        scv.tl.velocity_clusters(adata)
        scv.pl.scatter(adata, color='velocity_clusters')

    .. image:: https://user-images.githubusercontent.com/31883718/69625627-484dc480-1047-11ea-847f-6607a3430427.png
       :width: 600px


    Arguments
    ----------
    data : :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Key of velocities computed in `tl.velocity`
    match_with : `int`, optional (default: 100)
        The number of genes that appear in the returned tables.
    match_with: `str` (default: `'clusters'`)
        Match the names of the velocity clusters with the names of this key (.obs).
    sort_by: `str` or `None` (default: `'dpt_pseudotime'`)
        Sort velocity clusters by this key (.obs).
    resolution: `float` (default: 0.7)
        Resolution for louvain modularity.
    min_likelihood: `float` between `0` and `1` or `None` (default: `None`)
        Only rank velocity of genes with a likelihood higher than min_likelihood.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to data.

    Returns
    -------
    velocity_clusters : `.obs`
        Clusters obtained from applying louvain modularity on velocity expression.
    """  # noqa E501

    adata = data.copy() if copy else data

    logg.info("computing velocity clusters", r=True)

    tmp_filter = ~np.isnan(adata.layers[vkey].sum(0))
    if f"{vkey}_genes" in adata.var.keys():
        tmp_filter &= np.array(adata.var[f"{vkey}_genes"].values, dtype=bool)

    if "unspliced" in adata.layers.keys():
        n_counts = (adata.layers["unspliced"] > 0).sum(0)
        n_counts = n_counts.A1 if issparse(adata.layers["unspliced"]) else n_counts
        min_counts = min(50, np.percentile(n_counts, 50))
        tmp_filter &= np.ravel(n_counts > min_counts)

    if "r2" in adata.var.keys():
        r2 = adata.var.velocity_r2
        min_r2 = np.percentile(r2[r2 > 0], 50)
        tmp_filter &= r2 > min_r2

    if "dispersions_norm" in adata.var.keys():
        dispersions = adata.var.dispersions_norm
        min_dispersion = np.percentile(dispersions, 20)
        tmp_filter &= dispersions > min_dispersion

    if "fit_likelihood" in adata.var.keys() and min_likelihood is not None:
        tmp_filter &= adata.var["fit_likelihood"] > min_likelihood

    from anndata import AnnData

    vdata = AnnData(adata.layers[vkey][:, tmp_filter])
    vdata.obs = adata.obs.copy()
    vdata.var = adata.var[tmp_filter].copy()

    if "highly_variable" in vdata.var.keys():
        vdata.var["highly_variable"] = np.array(
            vdata.var["highly_variable"], dtype=bool
        )

    import scanpy as sc

    logg.switch_verbosity("off", module="scanpy")
    sc.pp.pca(vdata, n_comps=20, svd_solver="arpack")
    sc.pp.neighbors(vdata, n_pcs=20)
    sc.tl.louvain(vdata, resolution=0.7 if resolution is None else resolution)
    logg.switch_verbosity("on", module="scanpy")

    if sort_by == "velocity_pseudotime" and sort_by not in adata.obs.keys():
        velocity_pseudotime(adata, vkey=vkey)
    if sort_by in vdata.obs.keys():
        vc = vdata.obs["louvain"]
        vc_cats = vc.cat.categories
        mean_times = [np.mean(vdata.obs[sort_by][vc == cat]) for cat in vc_cats]
        vdata.obs["louvain"].cat.reorder_categories(
            vc_cats[np.argsort(mean_times)], inplace=True
        )

    if isinstance(match_with, str) and match_with in adata.obs.keys():
        from .utils import most_common_in_list

        vc = vdata.obs["louvain"]
        cats_nums = {cat: 0 for cat in adata.obs[match_with].cat.categories}
        for cat in vc.cat.categories:
            cells_in_cat = np.where(vc == cat)[0]
            new_cat = most_common_in_list(adata.obs[match_with][cells_in_cat])
            cats_nums[new_cat] += 1
            vc = vc.cat.rename_categories({cat: f"{new_cat} ({cats_nums[new_cat]})"})
        vdata.obs["louvain"] = vc
    else:
        vdata.obs["louvain"].cat.categories = np.arange(
            len(vdata.obs["louvain"].cat.categories)
        )
    adata.obs[f"{vkey}_clusters"] = vdata.obs["louvain"].copy()

    del vdata

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        f"    '{vkey}_clusters', "
        f"clusters based on louvain modularity on velocity vector field (adata.obs)"
    )

    return adata if copy else None


def rank_velocity_genes(
    data,
    vkey="velocity",
    n_genes=100,
    groupby=None,
    match_with=None,
    resolution=None,
    min_counts=None,
    min_r2=None,
    min_corr=None,
    min_dispersion=None,
    min_likelihood=None,
    copy=False,
):
    """Rank genes for velocity characterizing groups.

    This applies a differential expression test (Welch t-test with overestimated
    variance to be conservative) on velocity expression, to find genes in a cluster that
    show dynamics that is transcriptionally regulated differently compared to all other
    clusters (e.g. induction in that cluster and homeostasis in remaining population).
    If no clusters are given, it priorly computes velocity clusters by applying louvain
    modularity on velocity expression.

    .. code:: python

        scv.tl.rank_velocity_genes(adata, groupby='clusters')
        scv.pl.scatter(
            adata, basis=adata.uns['rank_velocity_genes']['names']['Beta'][:3]
        )
        pd.DataFrame(adata.uns['rank_velocity_genes']['names']).head()

    .. image:: https://user-images.githubusercontent.com/31883718/69626017-11c47980-1048-11ea-89f4-df3769df5ad5.png
       :width: 600px

    .. image:: https://user-images.githubusercontent.com/31883718/69626572-30774000-1049-11ea-871f-e8a30c42f10e.png
       :width: 600px

    Arguments
    ----------
    data : :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Key of velocities computed in `tl.velocity`
    n_genes : `int`, optional (default: 100)
        The number of genes that appear in the returned tables.
    groupby: `str`, `list` or `np.ndarray` (default: `None`)
        Key of observations grouping to consider.
    match_with: `str` or `None` (default: `None`)
        adata.obs key to separatively rank velocities on.
    resolution: `str` or `None` (default: `None`)
        Resolution for louvain modularity.
    min_counts: `float` (default: None)
        Minimum count of genes for consideration.
    min_r2: `float` (default: None)
        Minimum r2 value of genes for consideration.
    min_corr: `float` (default: None)
        Minimum Spearmans correlation coefficient between spliced and unspliced.
    min_dispersion: `float` (default: None)
        Minimum dispersion norm value of genes for consideration.
    min_likelihood: `float` between `0` and `1` or `None` (default: `None`)
        Only rank velocity of genes with a likelihood higher than min_likelihood.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to data.

    Returns
    -------
    rank_velocity_genes : `.uns`
        Structured array to be indexed by group id storing the gene
        names. Ordered according to scores.
    velocity_score : `.var`
        Storing the score for each gene for each group. Ordered according to scores.
    """  # noqa E501

    adata = data.copy() if copy else data

    if groupby is None or groupby == "velocity_clusters":
        velocity_clusters(
            adata,
            vkey=vkey,
            match_with=match_with,
            resolution=resolution,
            min_likelihood=min_likelihood,
        )
        groupby = f"{vkey}_clusters"

    logg.info("ranking velocity genes", r=True)

    if "spearmans_score" not in adata.var.keys():
        corr = vcorrcoef(
            np.array(adata.layers["Ms"]).T,
            np.array(adata.layers["Mu"].T),
            mode="spearmans",
        )
        adata.var["spearmans_score"] = np.clip(corr, 0, None)

    tmp_filter = ~np.isnan(adata.layers[vkey].sum(0))
    if f"{vkey}_genes" in adata.var.keys():
        tmp_filter &= np.array(adata.var[f"{vkey}_genes"].values, dtype=bool)

    if "unspliced" in adata.layers.keys():
        n_counts = (adata.layers["unspliced"] > 0).sum(0)
        n_counts = n_counts.A1 if issparse(adata.layers["unspliced"]) else n_counts
        min_counts = (
            min(50, np.percentile(n_counts, 50)) if min_counts is None else min_counts
        )
        tmp_filter &= np.ravel(n_counts > min_counts)

    if f"{vkey}_r2" in adata.var.keys():
        r2 = adata.var[f"{vkey}_r2"]
        min_r2 = 0.1 if min_r2 is None else min_r2  # np.percentile(r2[r2 > 0], 50)
        tmp_filter &= r2 > min_r2

    if "spearmans_score" in adata.var.keys():
        corr = adata.var["spearmans_score"]
        min_corr = (
            0.1 if min_corr is None else min_corr
        )  # np.percentile(r2[r2 > 0], 50)
        tmp_filter &= corr > min_corr

    if "dispersions_norm" in adata.var.keys():
        dispersions = adata.var.dispersions_norm
        min_dispersion = 0 if min_dispersion is None else min_dispersion
        tmp_filter &= dispersions > min_dispersion

    if "fit_likelihood" in adata.var.keys():
        fit_likelihood = adata.var["fit_likelihood"]
        min_likelihood = 0.1 if min_likelihood is None else min_likelihood
        tmp_filter &= fit_likelihood > min_likelihood

    X = adata[:, tmp_filter].layers[vkey]
    groups, groups_masks = select_groups(adata, key=groupby)

    n_groups = groups_masks.shape[0]
    sizes = groups_masks.sum(1)

    mean, var = np.zeros((n_groups, X.shape[1])), np.zeros((n_groups, X.shape[1]))
    for i, mask in enumerate(groups_masks):
        mean[i], var[i] = get_mean_var(X[mask])

    # test each against the union of all other groups
    rankings_gene_names, rankings_gene_scores = [], []
    for i in range(n_groups):
        mask_rest = ~groups_masks[i]
        mean_rest, var_rest = get_mean_var(X[mask_rest])
        size_rest = sizes[i]  # else mask_rest.sum() if method == 't-test'

        scores = (mean[i] - mean_rest) / np.sqrt(
            var[i] / sizes[i] + var_rest / size_rest
        )
        scores = np.nan_to_num(scores)

        # equivalent to but much faster than np.argsort(scores)[-10:]
        if n_genes > X.shape[1]:
            n_genes = X.shape[1]
        idx = np.argpartition(scores, -n_genes)[-n_genes:]
        idx = idx[np.argsort(scores[idx])[::-1]]

        rankings_gene_names.append(adata[:, tmp_filter].var_names[idx].values)
        rankings_gene_scores.append(scores[idx])

    rankings_gene_names = np.array([list(n) for n in rankings_gene_names])
    rankings_gene_scores = np.array([list(n) for n in rankings_gene_scores])

    all_names = rankings_gene_names.T.flatten()
    all_scores = rankings_gene_scores.T.flatten()
    vscore = np.zeros(adata.n_vars, dtype=int)
    for i, name in enumerate(adata.var_names):
        if name in all_names:
            vscore[i] = all_scores[np.where(name == all_names)[0][0]]
    adata.var["velocity_score"] = vscore

    key = "rank_velocity_genes"
    if key not in adata.uns.keys():
        adata.uns[key] = {}

    adata.uns[key] = {
        "names": np.rec.fromarrays(
            [n for n in rankings_gene_names], dtype=[(f"{rn}", "U50") for rn in groups]
        ),
        "scores": np.rec.fromarrays(
            [n.round(2) for n in rankings_gene_scores],
            dtype=[(f"{rn}", "float32") for rn in groups],
        ),
        "params": {
            "groupby": groupby,
            "reference": "rest",
            "method": "t-test_overestim_var",
            "use_raw": True,
        },
    }

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        f"    '{key}', sorted scores by group ids (adata.uns) \n"
        "    'spearmans_score', spearmans correlation scores (adata.var)"
    )

    return adata if copy else None



import warnings

import numpy as np
from scipy.sparse import csr_matrix, issparse

from scvelo import settings
from scvelo.preprocessing.neighbors import get_neighs
from scvelo.tools.transition_matrix import transition_matrix
from .docs import doc_params, doc_scatter
from .scatter import scatter
from .utils import (
    default_basis,
    default_size,
    get_basis,
    get_components,
    groups_to_bool,
    savefig_or_show,
)


@doc_params(scatter=doc_scatter)
def velocity_graph(
    adata,
    basis=None,
    vkey="velocity",
    which_graph=None,
    n_neighbors=10,
    arrows=None,
    arrowsize=3,
    alpha=0.8,
    perc=None,
    threshold=None,
    edge_width=0.2,
    edge_color="grey",
    edges_on_top=None,
    color=None,
    layer=None,
    size=None,
    groups=None,
    components=None,
    title=None,
    dpi=None,
    show=None,
    save=None,
    ax=None,
    **kwargs,
):
    """\
    Plot of the velocity graph.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    which_graph: `'velocity_graph'` or `'connectivities'`  (default: `None`)
        Whether to show transitions from velocity graph or neighbor connectivities.
    n_neighbors: `int` (default: 10)
        Number of neighbors to be included for generating connectivity / velocity graph.
    arrows: `bool` (default: `None`)
        Whether to display arrows instead of edges.
        Recommended to be used only on a cluster by setting groups parameter.
    arrowsize: `int` (default: 3)
        Size of the arrow heads.

    {scatter}

    Returns
    -------
    `matplotlib.Axis` if `show==False`
    """

    basis = default_basis(adata, **kwargs) if basis is None else get_basis(adata, basis)
    kwargs.update(
        {
            "basis": basis,
            "title": which_graph if title is None else title,
            "alpha": alpha,
            "components": components,
            "groups": groups,
            "dpi": dpi,
            "show": False,
            "save": None,
        }
    )
    ax = scatter(adata, layer=layer, color=color, size=size, ax=ax, zorder=0, **kwargs)

    from networkx import DiGraph, Graph

    if which_graph in {"neighbors", "connectivities"}:
        T = get_neighs(adata, "connectivities").copy()
        if perc is not None or threshold is not None:
            if threshold is None:
                threshold = np.percentile(T.data, perc)
            T.data[T.data < threshold] = 0
            T.eliminate_zeros()
    elif which_graph in adata.uns.keys():
        T = adata.uns[which_graph].copy()
        if perc is not None or threshold is not None:
            if threshold is None:
                threshold = np.percentile(T.data, perc)
            T.data[T.data < threshold] = 0
            T.eliminate_zeros()
    elif hasattr(adata, "obsp") and which_graph in adata.obsp.keys():
        T = adata.obsp[which_graph].copy()
        if perc is not None or threshold is not None:
            if threshold is None:
                threshold = np.percentile(T.data, perc)
            T.data[T.data < threshold] = 0
            T.eliminate_zeros()
    else:
        if threshold is None:
            threshold = 0.05
        T = transition_matrix(
            adata,
            vkey=vkey,
            weight_indirect_neighbors=0,
            n_neighbors=n_neighbors,
            perc=perc,
            threshold=threshold,
        )

    if groups is not None:
        if issparse(T):
            T = T.A
        T[~groups_to_bool(adata, groups, color)] = 0
        T = csr_matrix(T)
        T.eliminate_zeros()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X_emb = adata.obsm[f"X_{basis}"][:, get_components(components, basis)]
        node_size = (kwargs["size"] if "size" in kwargs else default_size(adata)) / 4
        edges = draw_networkx_edges(
            DiGraph(T) if arrows else Graph(T),
            X_emb,
            node_size=node_size,
            width=edge_width,
            edge_color=edge_color,
            arrowsize=arrowsize,
            ax=ax,
        )
        if not arrows and not edges_on_top:
            edges.set_zorder(-2)
            edges.set_rasterized(settings._vector_friendly)

    savefig_or_show(dpi=dpi, save=save, show=show)
    if show is False:
        return ax


def draw_networkx_edges(
    G,
    pos,
    edgelist=None,
    width=1.0,
    edge_color="k",
    style="solid",
    alpha=None,
    arrowstyle="-|>",
    arrowsize=3,
    edge_cmap=None,
    edge_vmin=None,
    edge_vmax=None,
    ax=None,
    arrows=True,
    label=None,
    node_size=300,
    nodelist=None,
    node_shape="o",
    connectionstyle=None,
    min_source_margin=0,
    min_target_margin=0,
):
    """Draw the edges of the graph G. Adjusted from networkx."""
    try:
        from numbers import Number

        import matplotlib.pyplot as plt
        from matplotlib.collections import LineCollection
        from matplotlib.colors import colorConverter, Colormap, Normalize
        from matplotlib.patches import FancyArrowPatch
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()

    if edgelist is None:
        edgelist = list(G.edges())

    if not edgelist or len(edgelist) == 0:  # no edges!
        return None

    if nodelist is None:
        nodelist = list(G.nodes())

    # FancyArrowPatch handles color=None different from LineCollection
    if edge_color is None:
        edge_color = "k"

    # set edge positions
    edge_pos = np.asarray([(pos[e[0]], pos[e[1]]) for e in edgelist])

    # Check if edge_color is an array of floats and map to edge_cmap.
    # This is the only case handled differently from matplotlib
    if (
        np.iterable(edge_color)
        and (len(edge_color) == len(edge_pos))
        and np.alltrue([isinstance(c, Number) for c in edge_color])
    ):
        if edge_cmap is not None:
            assert isinstance(edge_cmap, Colormap)
        else:
            edge_cmap = plt.get_cmap()
        if edge_vmin is None:
            edge_vmin = min(edge_color)
        if edge_vmax is None:
            edge_vmax = max(edge_color)
        color_normal = Normalize(vmin=edge_vmin, vmax=edge_vmax)
        edge_color = [edge_cmap(color_normal(e)) for e in edge_color]

    if not G.is_directed() or not arrows:
        edge_collection = LineCollection(
            edge_pos,
            colors=edge_color,
            linewidths=width,
            antialiaseds=(1,),
            linestyle=style,
            transOffset=ax.transData,
            alpha=alpha,
        )

        edge_collection.set_cmap(edge_cmap)
        edge_collection.set_clim(edge_vmin, edge_vmax)

        edge_collection.set_zorder(1)  # edges go behind nodes
        edge_collection.set_label(label)
        ax.add_collection(edge_collection)

        return edge_collection

    arrow_collection = None

    if G.is_directed() and arrows:
        # Note: Waiting for someone to implement arrow to intersection with
        # marker.  Meanwhile, this works well for polygons with more than 4
        # sides and circle.

        def to_marker_edge(marker_size, marker):
            if marker in "s^>v<d":  # `large` markers need extra space
                return np.sqrt(2 * marker_size) / 2
            else:
                return np.sqrt(marker_size) / 2

        # Draw arrows with `matplotlib.patches.FancyarrowPatch`
        arrow_collection = []
        mutation_scale = arrowsize  # scale factor of arrow head

        # FancyArrowPatch doesn't handle color strings
        arrow_colors = colorConverter.to_rgba_array(edge_color, alpha)
        for i, (src, dst) in enumerate(edge_pos):
            x1, y1 = src
            x2, y2 = dst
            shrink_source = 0  # space from source to tail
            shrink_target = 0  # space from  head to target
            if np.iterable(node_size):  # many node sizes
                source, target = edgelist[i][:2]
                source_node_size = node_size[nodelist.index(source)]
                target_node_size = node_size[nodelist.index(target)]
                shrink_source = to_marker_edge(source_node_size, node_shape)
                shrink_target = to_marker_edge(target_node_size, node_shape)
            else:
                shrink_source = shrink_target = to_marker_edge(node_size, node_shape)

            if shrink_source < min_source_margin:
                shrink_source = min_source_margin

            if shrink_target < min_target_margin:
                shrink_target = min_target_margin

            if len(arrow_colors) == len(edge_pos):
                arrow_color = arrow_colors[i]
            elif len(arrow_colors) == 1:
                arrow_color = arrow_colors[0]
            else:  # Cycle through colors
                arrow_color = arrow_colors[i % len(arrow_colors)]

            if np.iterable(width):
                if len(width) == len(edge_pos):
                    line_width = width[i]
                else:
                    line_width = width[i % len(width)]
            else:
                line_width = width

            arrow = FancyArrowPatch(
                (x1, y1),
                (x2, y2),
                arrowstyle=arrowstyle,
                shrinkA=shrink_source,
                shrinkB=shrink_target,
                mutation_scale=mutation_scale,
                color=arrow_color,
                linewidth=line_width,
                connectionstyle=connectionstyle,
                linestyle=style,
                zorder=1,
            )  # arrows go behind nodes

            # There seems to be a bug in matplotlib to make collections of
            # FancyArrowPatch instances. Until fixed, the patches are added
            # individually to the axes instance.
            arrow_collection.append(arrow)
            ax.add_patch(arrow)

    # update view
    minx = np.amin(np.ravel(edge_pos[:, :, 0]))
    maxx = np.amax(np.ravel(edge_pos[:, :, 0]))
    miny = np.amin(np.ravel(edge_pos[:, :, 1]))
    maxy = np.amax(np.ravel(edge_pos[:, :, 1]))

    w = maxx - minx
    h = maxy - miny
    padx, pady = 0.05 * w, 0.05 * h
    corners = (minx - padx, miny - pady), (maxx + padx, maxy + pady)
    ax.update_datalim(corners)
    ax.autoscale_view()

    ax.tick_params(
        axis="both",
        which="both",
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False,
    )

    return arrow_collection



import numpy as np

import matplotlib.pyplot as pl
from matplotlib import rcParams

from scvelo.core import SplicingDynamics
from scvelo.tools.dynamical_model_utils import get_vars, tau_inv, unspliced, vectorize
from .utils import make_dense


def get_dynamics(adata, key="fit", extrapolate=False, sorted=False, t=None):
    alpha, beta, gamma, scaling, t_ = get_vars(adata, key=key)
    if extrapolate:
        u0_ = unspliced(t_, 0, alpha, beta)
        tmax = t_ + tau_inv(u0_ * 1e-4, u0=u0_, alpha=0, beta=beta)
        t = np.concatenate(
            [np.linspace(0, t_, num=500), t_ + np.linspace(0, tmax, num=500)]
        )
    elif t is None or t is True:
        t = adata.obs[f"{key}_t"].values if key == "true" else adata.layers[f"{key}_t"]

    tau, alpha, u0, s0 = vectorize(np.sort(t) if sorted else t, t_, alpha, beta, gamma)
    ut, st = SplicingDynamics(
        alpha=alpha, beta=beta, gamma=gamma, initial_state=[u0, s0]
    ).get_solution(tau)
    return alpha, ut, st


def compute_dynamics(
    adata, basis, key="true", extrapolate=None, sort=True, t_=None, t=None
):
    idx = adata.var_names.get_loc(basis) if isinstance(basis, str) else basis
    key = "fit" if f"{key}_gamma" not in adata.var_keys() else key
    alpha, beta, gamma, scaling, t_ = get_vars(adata[:, basis], key=key)

    if "fit_u0" in adata.var.keys():
        u0_offset, s0_offset = adata.var["fit_u0"][idx], adata.var["fit_s0"][idx]
    else:
        u0_offset, s0_offset = 0, 0

    if t is None or isinstance(t, bool) or len(t) < adata.n_obs:
        t = (
            adata.obs[f"{key}_t"].values
            if key == "true"
            else adata.layers[f"{key}_t"][:, idx]
        )

    if extrapolate:
        u0_ = unspliced(t_, 0, alpha, beta)
        tmax = np.max(t) if True else tau_inv(u0_ * 1e-4, u0=u0_, alpha=0, beta=beta)
        t = np.concatenate(
            [np.linspace(0, t_, num=500), np.linspace(t_, tmax, num=500)]
        )

    tau, alpha, u0, s0 = vectorize(np.sort(t) if sort else t, t_, alpha, beta, gamma)

    ut, st = SplicingDynamics(
        alpha=alpha, beta=beta, gamma=gamma, initial_state=[u0, s0]
    ).get_solution(tau, stacked=False)
    ut, st = ut * scaling + u0_offset, st + s0_offset
    return alpha, ut, st


def show_full_dynamics(
    adata,
    basis,
    key="true",
    use_raw=False,
    linewidth=1,
    linecolor=None,
    show_assignments=None,
    ax=None,
):
    if ax is None:
        ax = pl.gca()
    color = linecolor if linecolor else "grey" if key == "true" else "purple"
    linewidth = 0.5 * linewidth if key == "true" else linewidth
    label = "learned dynamics" if key == "fit" else "true dynamics"
    line = None

    if key != "true":
        _, ut, st = compute_dynamics(
            adata, basis, key, extrapolate=False, sort=False, t=show_assignments
        )
        if not isinstance(show_assignments, str) or show_assignments != "only":
            ax.scatter(st, ut, color=color, s=1)
        if show_assignments is not None and show_assignments is not False:
            skey, ukey = (
                ("spliced", "unspliced")
                if use_raw or "Ms" not in adata.layers.keys()
                else ("Ms", "Mu")
            )
            s, u = (
                make_dense(adata[:, basis].layers[skey]).flatten(),
                make_dense(adata[:, basis].layers[ukey]).flatten(),
            )
            ax.plot(
                np.array([s, st]),
                np.array([u, ut]),
                color="grey",
                linewidth=0.1 * linewidth,
            )

    if not isinstance(show_assignments, str) or show_assignments != "only":
        _, ut, st = compute_dynamics(
            adata, basis, key, extrapolate=True, t=show_assignments
        )
        (line,) = ax.plot(st, ut, color=color, linewidth=linewidth, label=label)

        idx = adata.var_names.get_loc(basis)
        beta, gamma = adata.var[f"{key}_beta"][idx], adata.var[f"{key}_gamma"][idx]
        xnew = np.linspace(np.min(st), np.max(st))
        ynew = gamma / beta * (xnew - np.min(xnew)) + np.min(ut)
        ax.plot(xnew, ynew, color=color, linestyle="--", linewidth=linewidth)
    return line, label


def simulation(
    adata,
    var_names="all",
    legend_loc="upper right",
    legend_fontsize=20,
    linewidth=None,
    dpi=None,
    xkey="true_t",
    ykey=None,
    colors=None,
    **kwargs,
):
    from scvelo.tools.utils import make_dense
    from .scatter import scatter

    if ykey is None:
        ykey = ["unspliced", "spliced", "alpha"]
    if colors is None:
        colors = ["darkblue", "darkgreen", "grey"]
    var_names = (
        adata.var_names
        if isinstance(var_names, str) and var_names == "all"
        else [name for name in var_names if name in adata.var_names]
    )

    figsize = rcParams["figure.figsize"]
    ncols = len(var_names)
    for i, gs in enumerate(
        pl.GridSpec(
            1, ncols, pl.figure(None, (figsize[0] * ncols, figsize[1]), dpi=dpi)
        )
    ):
        idx = adata.var_names.get_loc(var_names[i])
        alpha, ut, st = compute_dynamics(adata, idx)
        t = (
            adata.obs[xkey]
            if xkey in adata.obs.keys()
            else make_dense(adata.layers["fit_t"][:, idx])
        )
        idx_sorted = np.argsort(t)
        t = t[idx_sorted]

        ax = pl.subplot(gs)
        _kwargs = {"alpha": 0.3, "title": "", "xlabel": "time", "ylabel": "counts"}
        _kwargs.update(kwargs)
        linewidth = 1 if linewidth is None else linewidth

        ykey = [ykey] if isinstance(ykey, str) else ykey
        for j, key in enumerate(ykey):
            if key in adata.layers:
                y = make_dense(adata.layers[key][:, idx])[idx_sorted]
                ax = scatter(x=t, y=y, color=colors[j], ax=ax, show=False, **_kwargs)

            if key == "unspliced":
                ax.plot(t, ut, label="unspliced", color=colors[j], linewidth=linewidth)
            elif key == "spliced":
                ax.plot(t, st, label="spliced", color=colors[j], linewidth=linewidth)
            elif key == "alpha":
                largs = dict(linewidth=linewidth, linestyle="--")
                ax.plot(t, alpha, label="alpha", color=colors[j], **largs)

        pl.xlim(0)
        pl.ylim(0)
        if legend_loc != "none" and i == ncols - 1:
            pl.legend(loc=legend_loc, fontsize=legend_fontsize)



"""Shared docstrings for plotting function parameters.
"""
from textwrap import dedent


def doc_params(**kwds):
    """\
    Docstrings should start with "\" in the first line for proper formatting.
    """

    def dec(obj):
        obj.__doc__ = dedent(obj.__doc__).format(**kwds)
        return obj

    return dec


doc_scatter = """\
basis: `str` or list of `str` (default: `None`)
    Key for embedding. If not specified, use 'umap', 'tsne' or 'pca' (ordered by
    preference).
vkey: `str` or list of `str` (default: `None`)
    Key for velocity / steady-state ratio to be visualized.
color: `str`,  list of `str` or `None` (default: `None`)
    Key for annotations of observations/cells or variables/genes
use_raw : `bool` (default: `None`)
    Use `raw` attribute of `adata` if present.
layer: `str`,  list of `str` or `None` (default: `None`)
    Specify the layer for `color`.
color_map: `str` (default: `matplotlib.rcParams['image.cmap']`)
    String denoting matplotlib color map.
colorbar: `bool` (default: `False`)
    Whether to show colorbar.
palette: list of `str` (default: `None`)
    Colors to use for plotting groups (categorical annotation).
size: `float` (default: 5)
    Point size.
alpha: `float` (default: 1)
    Set blending - 0 transparent to 1 opaque.
linewidth: `float` (default: 1)
    Scaling factor for the width of occurring lines.
linecolor: `str` ir list of `str` (default: 'k')
    Color of lines from velocity fits, linear fits and polynomial fits
perc: tuple, e.g. [2,98] (default: `None`)
    Specify percentile for continuous coloring.
groups: `str` or list of `str` (default: `all groups`)
    Restrict to a few categories in categorical observation annotation.
    Multiple categories can be passed as list with ['cluster_1', 'cluster_3'],
    or as string with 'cluster_1, cluster_3'.
sort_order: `bool` (default: `True`)
    For continuous annotations used as color parameter,
    plot data points with higher values on top of others.
components: `str` or list of `str` (default: '1,2')
    For instance, ['1,2', '2,3'].
projection: {'2d', '3d'} (default: '2d')
    Projection of plot.
legend_loc: str (default: 'none')
    Location of legend, either 'on data', 'right margin' or valid keywords
    for matplotlib.legend.
legend_fontsize: `int` (default: `None`)
    Legend font size.
legend_fontweight: {'normal', 'bold', ...} (default: `None`)
    Legend font weight. A numeric value in range 0-1000 or a string.
    Defaults to 'bold' if `legend_loc = 'on data'`, otherwise to 'normal'.
    Available are `['light', 'normal', 'medium', 'semibold', 'bold', 'heavy', 'black']`.
legend_fontoutline: float (default: `None`)
    Line width of the legend font outline in pt. Draws a white outline using
    the path effect :class:`~matplotlib.patheffects.withStroke`.
legend_align_text: bool or str (default: `None`)
    Aligns the positions of the legend texts. Set the axis along which the best
    alignment should be determined. This can be 'y' or True (vertically),
    'x' (horizontally), or 'xy' (best alignment in both directions).
right_margin: `float` or list of `float` (default: `None`)
    Adjust the width of the space right of each plotting panel.
left_margin: `float` or list of `float` (default: `None`)
    Adjust the width of the space left of each plotting panel.
xlabel: `str` (default: `None`)
    Label of x-axis.
ylabel: `str` (default: `None`)
    Label of y-axis.
title: `str` (default: `None`)
    Provide title for panels either as, e.g. `["title1", "title2", ...]`.
fontsize: `float` (default: `None`)
    Label font size.
figsize: tuple (default: `(7,5)`)
    Figure size.
xlim: tuple, e.g. [0,1] or `None` (default: `None`)
    Restrict x-limits of the axis.
ylim: tuple, e.g. [0,1] or `None` (default: `None`)
    Restrict y-limits of the axis.
add_density: `bool` or `str` or `None` (default: `None`)
    Whether to show density of values along x and y axes.
    Color of the density plot can also be passed as `str`.
add_assignments: `bool` or `str` or `None` (default: `None`)
    Whether to add assignments to the model curve.
    Color of the assignments can also be passed as `str`.
add_linfit: `bool` or `str` or `None` (default: `None`)
    Whether to add linear regression fit to the data points.
    Color of the line can also be passed as `str`.
    Fitting with or without an intercept by passing `'intercept'` or `'no_intercept'`.
    A colored regression line with intercept is obtained with `'intercept, blue'`.
add_polyfit: `bool` or `str` or `int` or `None` (default: `None`)
    Whether to add polynomial fit to the data points. Color of the polyfit plot can also
    be passed as `str`. The degree of the polynomial fit can be passed as `int`
    (default is 2 for quadratic fit).
    Fitting with or without an intercept by passing `'intercept'` or `'no_intercept'`.
    A colored regression line with intercept is obtained with `'intercept, blue'`.
add_rug: `str` or `None` (default: `None`)
    If categorical observation annotation (e.g. 'clusters') is given, a rugplot is
    attached to the x-axis showing the data membership to each of the categories.
add_text: `str` (default: `None`)
    Text to be added to the plot, passed as `str`.
add_text_pos: `tuple`, e.g. [0.05, 0.95] (defaut: `[0.05, 0.95]`)
    Text position. Default is `[0.05, 0.95]`, positioning the text at top right corner.
add_margin: `float` (default: `None`)
    A value between [-1, 1] to add (positive) and reduce (negative) figure margins.
add_outline: `bool` or `str` (default: `False`)
    Whether to show an outline around scatter plot dots.
    Alternatively a string of cluster names can be passed, e.g. 'cluster_1, clusters_3'.
outline_width: tuple type `scalar` or `None` (default: `(0.3, 0.05)`)
    Width of the inner and outer outline
outline_color: tuple of type `str` or `None` (default: `('black', 'white')`)
    Inner and outer matplotlib color of the outline
n_convolve: `int` or `None` (default: `None`)
    If `int` is given, data is smoothed by convolution
    along the x-axis with kernel size `n_convolve`.
smooth: `bool` or `int` (default: `None`)
    Whether to convolve/average the color values over the nearest neighbors.
    If `int`, it specifies number of neighbors.
normalize_data: `bool` (default: `None`)
    Whether to rescale values for x, y to [0,1].
rescale_color: `tuple` (default: `None`)
    Boundaries for color rescaling, e.g. [0, 1], setting min/max values of the colorbar.
color_gradients: `str` or `np.ndarray` (default: `None`)
    Key for `.obsm` or array with color gradients by categories.
dpi: `int` (default: 80)
    Figure dpi.
frameon: `bool` (default: `True`)
    Draw a frame around the scatter plot.
ncols: `int` (default: `None`)
    Number of panels per row.
nrows: `int` (default: `None`)
    Number of panels per column.
wspace : `float` (default: None)
    Adjust the width of the space between multiple panels.
hspace : `float` (default: None)
    Adjust the height of the space between multiple panels.
show: `bool`, optional (default: `None`)
    Show the plot, do not return axis.
save: `bool` or `str`, optional (default: `None`)
    If `True` or a `str`, save the figure. A string is appended to the default filename.
    Infer the filetype if ending on {'.pdf', '.png', '.svg'}.
ax: `matplotlib.Axes`, optional (default: `None`)
    A matplotlib axes object. Only works if plotting a single component.\
"""



from ._anndata import (
    clean_obs_names,
    cleanup,
    get_df,
    get_initial_size,
    get_modality,
    get_size,
    make_dense,
    make_sparse,
    merge,
    set_initial_size,
    set_modality,
    show_proportions,
)
from ._arithmetic import clipped_log, invert, prod_sum, sum
from ._linear_models import LinearRegression
from ._metrics import l2_norm
from ._models import SplicingDynamics
from ._parallelize import get_n_jobs, parallelize

__all__ = [
    "clean_obs_names",
    "cleanup",
    "clipped_log",
    "get_df",
    "get_initial_size",
    "get_modality",
    "get_n_jobs",
    "get_size",
    "invert",
    "l2_norm",
    "LinearRegression",
    "make_dense",
    "make_sparse",
    "merge",
    "parallelize",
    "prod_sum",
    "set_initial_size",
    "set_modality",
    "show_proportions",
    "SplicingDynamics",
    "sum",
]



from scvelo.core import (
    clean_obs_names,
    cleanup,
    get_initial_size,
    merge,
    set_initial_size,
    show_proportions,
)
from scvelo.plotting.simulation import compute_dynamics
from scvelo.plotting.utils import (
    clip,
    interpret_colorkey,
    is_categorical,
    rgb_custom_colormap,
)
from scvelo.plotting.velocity_embedding_grid import compute_velocity_on_grid
from scvelo.preprocessing.moments import get_moments
from scvelo.preprocessing.neighbors import get_connectivities
from scvelo.read_load import (
    convert_to_ensembl,
    convert_to_gene_names,
    gene_info,
    load_biomart,
)
from scvelo.tools.optimization import get_weight, leastsq
from scvelo.tools.rank_velocity_genes import get_mean_var
from scvelo.tools.run import convert_to_adata, convert_to_loom
from scvelo.tools.score_genes_cell_cycle import get_phase_marker_genes
from scvelo.tools.transition_matrix import get_cell_transitions
from scvelo.tools.transition_matrix import transition_matrix as get_transition_matrix
from scvelo.tools.utils import *  # noqa
from scvelo.tools.velocity_graph import vals_to_csr

__all__ = [
    "cleanup",
    "clean_obs_names",
    "clip",
    "compute_dynamics",
    "compute_velocity_on_grid",
    "convert_to_adata",
    "convert_to_ensembl",
    "convert_to_gene_names",
    "convert_to_loom",
    "gene_info",
    "get_cell_transitions",
    "get_connectivities",
    "get_initial_size",
    "get_mean_var",
    "get_moments",
    "get_phase_marker_genes",
    "get_transition_matrix",
    "get_weight",
    "interpret_colorkey",
    "is_categorical",
    "leastsq",
    "load_biomart",
    "merge",
    "rgb_custom_colormap",
    "set_initial_size",
    "show_proportions",
    "vals_to_csr",
]



import numpy as np

import scvelo as scv


def test_einsum():
    from scvelo.tools.utils import norm, prod_sum_obs, prod_sum_var

    Ms, Mu = np.random.rand(5, 4), np.random.rand(5, 4)
    assert np.allclose(prod_sum_obs(Ms, Mu), np.sum(Ms * Mu, 0))
    assert np.allclose(prod_sum_var(Ms, Mu), np.sum(Ms * Mu, 1))
    assert np.allclose(norm(Ms), np.linalg.norm(Ms, axis=1))


def test_neighbors():
    adata = scv.datasets.simulation(random_seed=0, n_vars=100)
    scv.pp.filter_and_normalize(adata)
    scv.pp.pca(adata)
    scv.pp.neighbors(adata)
    adata_ = scv.pp.neighbors(adata, method="sklearn", copy=True)
    dists = np.round(adata.obsp["distances"][0].data, 2)
    dists_ = np.round(adata_.obsp["distances"][0].data, 2)
    assert np.all(dists == dists_)


def test_dynamical_model():
    adata = scv.datasets.simulation(random_seed=0, n_vars=10)
    scv.pp.filter_and_normalize(adata)
    scv.pp.moments(adata)
    scv.tl.recover_dynamics(adata, var_names=adata.var_names[0])
    assert np.round(adata[:, adata.var_names[0]].var["fit_alpha"][0], 4) == 4.7409


def test_pipeline():
    adata = scv.datasets.simulation(random_seed=0, n_vars=10)

    scv.pp.filter_and_normalize(adata, n_top_genes=5)
    scv.pp.pca(adata)
    scv.pp.moments(adata)

    scv.tl.recover_dynamics(adata)
    scv.tl.velocity(adata)
    scv.tl.velocity(adata, vkey="dynamical_velocity", mode="dynamical")
    adata.var.velocity_genes = True

    scv.tl.velocity_graph(adata)
    scv.tl.velocity_embedding(adata)

    scv.tl.velocity_confidence(adata)
    scv.tl.latent_time(adata)

    with scv.GridSpec() as pl:
        pl.velocity_graph(adata)
        pl.velocity_embedding(adata, arrow_length=3, arrow_size=3, c="latent_time")
        pl.velocity_embedding_grid(adata, scale=0.5, c="latent_time", cmap="gnuplot")
        pl.velocity_embedding_stream(adata, c=adata.var_names[0], layer="velocity")
        pl.scatter(adata, basis=adata.var_names[0], c="velocity", use_raw=True)
        pl.hist([adata.obs.initial_size_spliced, adata.obs.initial_size_unspliced])

    Ms, Mu = adata.layers["Ms"][0], adata.layers["Mu"][0]
    Vs, Vd = adata.layers["velocity"][0], adata.layers["dynamical_velocity"][0]
    Vgraph = adata.uns["velocity_graph"].data[:5]
    pars = adata[:, 0].var[["fit_alpha", "fit_gamma"]].values

    assert np.allclose(Ms, [0.8269, 1.0772, 0.9396, 1.0846, 1.0398], rtol=1e-2)
    assert np.allclose(Mu, [3.8412, 3.1976, 3.5523, 3.3433, 3.8006], rtol=1e-2)
    assert np.allclose(adata.X[0], [0.0, 0.0, 0.0, 0.4981, 0.0], rtol=1e-2)
    # assert np.allclose(Vpca, [0.0163, 0.0185, 0.0472, 0.0025], rtol=1e-2)
    assert np.allclose(Vd, [1.7582, 2.0214, 1.73, 0.6615, 1.5118], rtol=1e-2)
    assert np.allclose(Vs, [3.2961, 2.5254, 2.9926, 2.634, 3.1352], rtol=1e-2)
    assert np.allclose(Vgraph, [0.915, 0.5997, 0.8494, 0.1615, 0.7708], rtol=1e-2)
    assert np.allclose(pars, [4.9257, 0.3239], rtol=1e-2)


def test_highly_variable_subset():
    adata = scv.datasets.simulation(random_seed=0, n_vars=10)
    bdata = adata.copy()

    scv.pp.filter_and_normalize(adata, n_top_genes=5, subset_highly_variable=True)
    scv.pp.filter_and_normalize(bdata, n_top_genes=5, subset_highly_variable=False)

    scv.pp.pca(adata)
    scv.pp.pca(bdata)

    scv.pp.moments(adata, use_rep="pca")
    scv.pp.moments(bdata, use_rep="pca")

    scv.tl.velocity_graph(adata)
    scv.tl.velocity_graph(bdata)

    bdata._inplace_subset_var(bdata.var["highly_variable"])

    assert np.allclose(adata.layers["Ms"][0], bdata.layers["Ms"][0])
    assert np.allclose(adata.layers["velocity"][0], bdata.layers["velocity"][0])
    assert np.allclose(
        adata.uns["velocity_graph"].data[:5], bdata.uns["velocity_graph"].data[:5]
    )



from . import velocity
from .utils import make_unique_list
from .. import logging as logg
from .dynamical_model_utils import get_divergence
from .dynamical_model import DynamicsRecovery, write_pars, read_pars, recover_dynamics
import numpy as np
from scanpy.tools import louvain
from scipy.stats.distributions import laplace
from scipy.stats import spearmanr
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit


def get_orth_dists(S, U, beta, std_s, std_u, scaling):
    S_real = np.array((S + (beta * U)) / (1 + beta ** 2))
    orth_dists = np.array(S - S_real) / std_s
    orth_distu = np.array(U - (beta * S_real)) / std_u * scaling
    orth_distx = orth_distu ** 2 + orth_dists ** 2
    return orth_distx


def get_orth_fit(U, S, weights):
    U, S = U * weights.astype(int), S * weights.astype(int)
    a, b = np.sum(S * U, axis=0), np.sum(U ** 2 - S ** 2, axis=0)
    orth_beta = (b + ((b ** 2 + 4 * a ** 2) ** .5)) / (2 * a)
    return orth_beta


def appendInt(num):
    if num > 9:
        secondToLastDigit = str(num)[-2]
        if secondToLastDigit == '1':
            return 'th'
    lastDigit = num % 10
    if lastDigit == 1:
        return 'st'
    elif lastDigit == 2:
        return 'nd'
    elif lastDigit == 3:
        return 'rd'
    else:
        return 'th'


def expfun(x, a, b):
    return a * (-np.exp(-b * x) + 1)


def f(a, b):
    def fun(x):
        return a * (-np.exp(-b * x) + 1)

    return fun


def inverse(x, a, b):
    return -np.log((-x / a) + 1) / b


def get_curve_fit(X, Y, sigma):
    kwargs = dict(bounds=([0, 0], [np.inf, 1.5]), sigma=sigma, method='trf', maxfev=4000)
    fun, fup = None, None
    try:
        pars, cov = curve_fit(f=expfun, xdata=X, ydata=Y, **kwargs)
        up_dist = np.sum((Y - expfun(X, *pars)) ** 2)
        fun, fup = f(*pars), True
        x = np.linspace(0, np.max(X), 100)
        y = expfun(x, *pars)
    except RuntimeError:
        up_dist = np.inf
    try:
        down_pars, cov = curve_fit(f=expfun, xdata=Y, ydata=X, **kwargs)
        down_dist = np.sum((X - expfun(Y, *down_pars)) ** 2)
        if down_dist < up_dist:
            fun, fup = f(*down_pars), False
            y = np.linspace(0, np.max(Y), 100)
            x = expfun(y, *down_pars)
    except RuntimeError:
        down_dist = np.inf
    return fun, fup#, x, y  # np.min([down_dist, up_dist])*2#, y, x


def plot(S, U, fit_weights, weights, i, x2, y2, sigma=None):
    fig, ax = plt.subplots()
    ax.scatter(S[:, i], U[:, i], c="black", s=10)
    U, S = U * fit_weights.astype(int), S.copy() * fit_weights.astype(int)
    ax.scatter(S[:, i], U[:, i], c="orange", s=2)
    U, S = U * weights.astype(int), S * weights.astype(int)
    ax.scatter(S[:, i], U[:, i], c="yellow", s=4)
    ax.plot(x2, y2, "r")
    #if sigma != None:
    #    print(sigma[:,i])
    #ax.scatter(S[:, i], U[:, i], c=sigma[:,i], s=5)
    plt.show()


def differential_kinetic_test(adata, var_names=None, cluster_name="clusters", key="fit", fit_type="line",
                              test_with_refit=False,
                              min_cluster_cells=15, use_raw=False, refit_main_dynamics=True, refit_at_end=True,
                              **kwargs):
    """
    Parameters
    ----------
    adata
    var_names: list of "str"
        Names of variables/genes to use for the fitting.
        If none given, will use all genes having dynamical model fit.
    cluster_name: "str" (default: "clusters")
        Name of the adata.obs column by which to cluster the cells.
    key: "str" (default "fit")
        key under which the dynamical model fit parameters are saved
    fit_type: str "line" or "curve"
        line or curve fit to test if additional clusters also differential (in (2))
    test_with_refit: bool (default False)
        whether to fit the dyn-model without the cluster before testing for differential kinetic of that
        cluster (in (1))
    min_cluster_cells: int (default 15)
        Minimal number of cells needed in cluster.
    use_raw: bool (default False)
        Whether to use raw unspliced and spliced counts or first order moments.
    refit_main_dynamics: bool (default: True)
        Whether the kinetic fit should be recomputed to exclude the competing kinetic.
        with **kwargs optional additional parameters for the recover_dynamics function
    refit_at_end: bool (default: False)
        Whether the main kinetic fit should be recomputed at all iteration or only at the end when all competing
         kinetics have been identified. Only relevant if nk>1.
    Returns
    -------
    """
    # check if params are ok or set if needed
    if var_names is not None:
        var_names = np.array(
            [name for name in make_unique_list(var_names, allow_array=True) if name in adata.var_names])
        if len(var_names) == 0:
            raise ValueError('Variable name not found in var keys.')
    else:
        var_names = adata.var_names
    var_names = adata[:, var_names].var_names[~np.isnan(adata[:, var_names].var[key + "_alpha"])]
    if len(var_names) == 0:
        logg.error(
            "No dynamical model parameters found for key \"" + key + "\" in adata. Please call scv.tl.recover_dynamics "
                                                                     "before testing for differential kinetics.")
    if cluster_name not in adata.obs.keys():
        logg.error("Cluster name: " + cluster_name + " not in adata.obs.keys() \nUsing louvain clusters instead.")
        if "louvain" not in adata.obs.keys():
            louvain(adata)
        cluster_name = "louvain"
    if min_cluster_cells < 15:
        logg.warn("Normal approximation for the statistical test works only if there are enough points in the tested "
                  "cluster. Setting \"min_cluster_cells\" parameter to " + min_cluster_cells + "might cause inaccurate "
                                                                                               "results.")
    #########
    # fetch #
    #########
    idx = np.where([i in var_names for i in adata.var_names])[0]
    var_names = adata.var_names[idx]  # sort var_names according to adata order
    spliced, unspliced = "spliced" if use_raw else "Ms", "unspliced" if use_raw else "Mu"
    S, U = adata.layers[spliced][:, idx], adata.layers[unspliced][:, idx]
    scaling = np.array(adata.var.loc[var_names, :][key + "_scaling"])
    U = U / scaling[None, :]
    std_s, std_u, varx = np.array(adata.var.loc[var_names, :][key + "_std_s"]), np.array(
        adata.var.loc[var_names, :][key + "_std_u"]), np.array(adata.var.loc[var_names, :][key + "_variance"])
    weights = ((S > 0) & (U > 0))
    weights &= ((S > np.nanmax(S, axis=0) / 5) | (U > np.nanmax(U, axis=0) / 5))
    # sub_idx because get_divergence returns obs * n_vars with fit_alpha, thus only subset amongst the vars with dynfit
    sub_idx = np.where([i in var_names for i in adata[:, ~np.isnan(adata.var[key + "_alpha"].values)].var_names])[0]
    dyn_distx = get_divergence(adata, mode='distx', use_connectivities=False, use_raw=use_raw)[:, sub_idx]
    clusters = adata.obs[cluster_name]
    clus = list(clusters.cat.categories)
    ###########
    # testing #
    ###########
    main_kinetic = np.ones(weights.shape).astype(bool)  # stores whether the cells are part of a competing kinetic
    all_diff = np.array(np.repeat("", adata.shape[1]), dtype=object)  # all competing kinetics

    LR, i, where, n = [], 0, [True] * len(var_names), len(var_names)
    while np.sum(where) > 0:
        # (1) get cluster that fits worst to main kinetic
        best_clus, best_beta, best_pval = np.array(np.repeat("", n), dtype=object), np.zeros(n), np.ones(n)
        for c in clus:
            if test_with_refit:
                bdata = adata.copy()
                bdata.var["fit_diff_kinetics"] = (cluster_name + ":" + c) if i == 0 else [cluster_name + ":" + i[1:] + str("," + c) for i in all_diff]
                #print(bdata.var["fit_diff_kinetics"])
                recover_dynamics(bdata, var_names[where],
                                 use_raw=use_raw, add_key=key, load_pars=False, main=True, **kwargs)
                # var_names[where]
                dyn_distx = get_divergence(bdata, mode='distx', use_connectivities=False, use_raw=use_raw)[:, sub_idx]
            weights_cluster = np.array([(clusters == c).tolist()] * n).T & weights & main_kinetic
            orth_beta = get_orth_fit(U, S, weights=weights_cluster)  # include inner vals
            pval, lr = orth_diff_kinetics(dyn_distx, weights_cluster, S, U, orth_beta, varx, std_s, std_u, scaling)
            if fit_type == "line":
                LR.extend(lr.tolist())
            test = (pval < best_pval) & (np.sum(weights_cluster, axis=0) > min_cluster_cells)
            best_pval[test], best_clus[test], best_beta[test] = pval[test], c, orth_beta[test]
        where = (best_pval < 0.05)
        logg.info(str(np.sum(where)) + " genes with " + str(i + 1) + " competing kinetic" + ("s." if i > 0 else "."))
        if np.sum(where) == 0:
            break
        main_kinetic[:, ~where], best_clus[~where] = False, ""  # no need to look again at genes with only main kinetic
        if test_with_refit:
            bdata.var["fit_diff_kinetics"] = [cluster_name + ":" + i if i != "" else "" for i in best_clus]
            recover_dynamics(bdata, var_names[where],
                             use_raw=use_raw, add_key=key, load_pars=False, main=True, **kwargs)
            dyn_distx = get_divergence(bdata, mode='distx', use_connectivities=False, use_raw=use_raw)[:, sub_idx]
        prev = (np.array([(clusters == j) for j in best_clus]).T & weights)
        # to save test parameters
        mn = main_kinetic[:, where]
        # (2) check if other cluster fit better to competing kinetic than to main one
        if fit_type == "line":
            diff_kinetics = best_clus[where]
            beta, best_pval = np.zeros(n), np.zeros(n)
            beta[:], best_pval[:] = np.nan, np.nan
            for c in clus:
                weights_cluster = (np.array(
                    [(clusters == c).tolist()] * len(var_names)).T & ~prev & weights & main_kinetic)[:,
                                  where]
                pval, lr = orth_diff_kinetics(dyn_distx[:, where], weights_cluster, S[:, where], U[:, where],
                                              best_beta[where], varx[where], std_s[where], std_u[where], scaling[where])
                diff_kinetics[pval < 0.05] += "," + c
            # get final pval
            best_clus[where] = diff_kinetics
            final_weights = np.array([[i in diff.split(",") for i in clusters] for diff in best_clus[where]]).T \
                            & (weights & main_kinetic)[:, where]
            mn[final_weights] = False
            pval_kinetics, lr = orth_diff_kinetics(dyn_distx[:, where], final_weights, S[:, where], U[:, where],
                                                   best_beta[where],
                                                   varx[where], std_s[where], std_u[where], scaling[where])
        elif fit_type == "curve":
            fit, m = where.copy(), 1
            while np.sum(fit) > 0:
                best_lr, b_pval, diff_kinetics = np.ones(np.sum(fit)) * -np.inf, np.ones(np.sum(fit)), best_clus[fit]
                for c in clus:
                    weights_cluster = ((np.array(
                        [(clusters == c).tolist()] * len(var_names)).T) & ~prev & weights & main_kinetic)[:,
                                      fit]
                    sigma = (weights_cluster.astype(int) * np.sum(weights_cluster, axis=0)) * m + (
                            prev[:, fit].astype(int) * np.sum(prev[:, fit], axis=0))  # m scales number of prev kinetics
                    pval, lr = curve_diff_kinetics(dyn_distx[:, fit], weights_cluster, (prev[:, fit] | weights_cluster),
                                                   sigma, (S / std_s)[:, fit], (U / std_u * scaling)[:, fit], varx[fit])
                    LR.extend(lr)
                    diff_kinetics[pval < b_pval] = c  # ,
                    b_pval[pval < b_pval] = pval[pval < b_pval]
                c = best_clus[fit]
                c[b_pval < 0.05] += "," + diff_kinetics[b_pval < 0.05]
                best_clus[fit], fit[fit] = c, b_pval < 0.05
                prev = (np.array([[i in j for i in clusters] for j in best_clus]).T & weights)
                m += 1
                # logg.info(str(np.sum(fit)) + " genes with " + str(m) + " clusters in the competing kinetic.")
            prev = (np.array([[i in j for i in clusters] for j in best_clus]).T & weights)
            pval_kinetics, lr = curve_diff_kinetics(dyn_distx[:, where], prev[:, where], prev[:, where],
                                                    prev[:, where], (S / std_s)[:, where],
                                                    (U / std_u * scaling)[:, where], varx[where])
            mn[prev[:, where]] = False
        main_kinetic[:, where] = mn
        nidx = idx[where]  # index of multi-kinetic genes
        best_pval = np.ones(adata.shape[1]) * np.nan
        best_pval[nidx], all_diff[nidx] = pval_kinetics, all_diff[nidx] + "," + best_clus[nidx]
        local_diff = np.array([cluster_name + ":" + best_clus[i] if i in nidx else "" for i in range(adata.shape[1])])

        # save diff kinetics in var
        main_diff = [cluster_name + ":" + d[1:] if d != "" else "" for d in all_diff]
        write_pars(adata, [main_diff], pars_names=["diff_kinetics"], add_key=key)
        write_pars(adata, [best_pval, local_diff],
                   pars_names=["pval_kinetics", "kinetics"], add_key=key + str(i + 1))
        if refit_main_dynamics:
            logg.warn("Refitting main dynamic for genes with competing kinetics.")
            # refit multi-kinetic genes where competing kinetic has been found at this iteration
            recover_dynamics(adata, adata[:, adata.var[key + str(i + 1) + "_kinetics"] != ""].var_names,
                             use_raw=use_raw, add_key=key, load_pars=False, main=True, **kwargs)
        i += 1
    if (not refit_main_dynamics) & refit_at_end:
        logg.warn("Refitting main dynamic for genes with competing kinetics.")
        recover_dynamics(adata, adata[:, adata.var[key + "_diff_kinetics"] != ""].var_names, use_raw=use_raw,
                         add_key=key, load_pars=False, main=True, **kwargs)
    return LR


def orth_diff_kinetics(distx, weights, S, U, orth_beta, varx, std_s, std_u, scaling):
    U, S = U * weights.astype(int), S * weights.astype(int)
    orth_distx = get_orth_dists(S, U, orth_beta, std_s, std_u, scaling)
    denom = varx * np.sqrt(8 * np.sum(weights, axis=0))  # sum of weights is all cells taken into account
    distx = distx * 0.5  # todo double check!
    lr = (np.sum(distx * weights.astype(int), axis=0) - np.sum(orth_distx * weights.astype(int), axis=0)) / denom
    return np.array([laplace.sf(x) for x in lr]), lr


def curve_diff_kinetics(distx, testing_weights, fitting_weights, sigma, S, U, varx):
    Ufit, Sfit = U.copy() * fitting_weights.astype(int), S.copy() * fitting_weights.astype(int)
    Ut, St = U.copy() * testing_weights.astype(int), S.copy() * testing_weights.astype(int)
    curve_dist = np.ones((U.shape[1])) * np.inf
    distx = np.sum(distx * testing_weights.astype(int), axis=0) * 0.5
    denom = varx * np.sqrt(8 * np.sum(testing_weights, axis=0))
    # for each gene, get curve fit
    for i in np.where(np.sum(testing_weights, axis=0) > 4)[0]:
        # get curve fit
        sig = (sigma[:, i] / np.sum(sigma[:, i])) * np.sum(fitting_weights[:, i])

        fun, fup = get_curve_fit(Sfit[:, i][sig > 0], Ufit[:, i][sig > 0], sigma=sig[sig > 0])
        if fup is None:  # curve fit did not converge
            curve_dist[i] = np.inf
        elif fup:
            curve_dist[i] = np.sum((Ut[:, i] - fun(St[:, i])) ** 2)
        else:
            curve_dist[i] = np.sum((St[:, i] - fun(Ut[:, i])) ** 2)
        # lr = (distx[i] - curve_dist[i]) / denom[i]
        # if laplace.sf(lr) < 0.05:
        #    print(lr)
        #    plt.hist(sig[sig > 0], bins=100)
        #    plot(S, U, fitting_weights, testing_weights, i, x, y)
    lr = (distx - curve_dist) / denom
    return np.array([laplace.sf(x) for x in lr]), lr


def recover_second_kinetic_velocity(adata, use_raw=False):  # just for rename not to break everything right away
    recover_competing_kinetic_velocity(adata, use_raw)


def recover_competing_kinetic_velocity(adata, use_raw=False, nk=None):
    logg.info("Recovering main velocity.")
    velocity(adata, mode='dynamical', diff_kinetics=True, use_raw=True)  # velocity=0 for clusters in fit_diff_kinetics
    nk = np.max([(int(i.split("_")[0][3:]) if len(i.split("_")[0][3:]) else 0) if i.startswith("fit") else 1
                 for i in adata.var.columns.tolist()]) if nk is None else nk  # get highest x for fitx_ in var

    for i in range(nk):
        logg.info("Recovering velocity of " + str(i + 1) + appendInt(i + 1) + " competing kinetic.")

        subdata = adata.copy()
        competing_kinetic = 'fit' + str(i + 1) + "_kinetics"
        subdata.var['fit_diff_kinetics'] = ""
        var_names = subdata[:, adata.var[competing_kinetic] != ""].var_names

        alpha, beta, gamma, t_, scaling, std_u, std_s, likelihood, u0, s0, pval, steady_u, steady_s, varx = read_pars(
            subdata)
        idx, L, P = [], [], []
        T = subdata.layers['fit_t'] if 'fit_t' in subdata.layers.keys() else np.zeros(subdata.shape) * np.nan
        Tau = subdata.layers['fit_tau'] if 'fit_tau' in subdata.layers.keys() else np.zeros(subdata.shape) * np.nan
        Tau_ = subdata.layers['fit_tau_'] if 'fit_tau_' in subdata.layers.keys() else np.zeros(subdata.shape) * np.nan
        diff_kinetics = ["" for x in range(subdata.n_vars)]

        progress = logg.ProgressReporter(len(var_names))
        for gene in var_names:
            progress.update()
            ix = np.where(subdata.var_names == gene)[0][0]
            diff_k = adata.var[competing_kinetic][ix].split(":")
            groupby = diff_k[0]
            clusters = subdata.obs[groupby if isinstance(groupby, str) else "clusters"]
            groups = clusters.cat.categories
            v = diff_k[1].split(",")
            m = [i in v for i in clusters]
            # check if cluster in steady-state
            s, u = adata.layers["spliced" if use_raw else "Ms"][m, ix], \
                   subdata.layers["unspliced" if use_raw else "Mu"][
                       m, ix]
            if spearmanr(s, u).pvalue < 0.01:  # 0.01:
                dm = DynamicsRecovery(subdata, gene, use_raw=use_raw, load_pars=False, model=m)
                if dm.recoverable:
                    dm.fit()
                    idx.append(ix)
                    T[:, ix], Tau[:, ix], Tau_[:, ix] = dm.t, dm.tau, dm.tau_
                    alpha[ix], beta[ix], gamma[ix], t_[ix], scaling[ix] = dm.pars[:, -1]
                    u0[ix], s0[ix], pval[ix], steady_u[ix], steady_s[
                        ix] = dm.u0, dm.s0, dm.pval_steady, dm.steady_u, dm.steady_s
                    beta[ix] /= scaling[ix]
                    steady_u[ix] *= scaling[ix]
                    std_u[ix], std_s[ix], likelihood[ix], varx[ix] = dm.std_u, dm.std_s, dm.likelihood, dm.varx
                    L.append(dm.loss)
                diff_kinetics[ix] = groupby + ":" + ",".join(groups[[i not in diff_k for i in groups]])
        progress.finish()

        write_pars(subdata,
                   [alpha, beta, gamma, t_, scaling, std_u, std_s, likelihood, u0, s0, pval, steady_u, steady_s, varx])
        write_pars(subdata, [diff_kinetics], pars_names=["diff_kinetics"])

        # write_pars(adata,
        #           [alpha, beta, gamma, t_, scaling, std_u, std_s, likelihood, u0, s0, pval, steady_u, steady_s, varx])

        subdata.layers['fit_t'], subdata.layers['fit_tau'], subdata.layers['fit_tau_'] = T, Tau, Tau_
        # now recover velocities
        velocity(subdata, mode='dynamical', diff_kinetics=True, use_raw=use_raw)
        # merge velocities
        where = subdata.var['fit_diff_kinetics'] != ""
        velo, velo_u = adata.layers["velocity"], adata.layers["velocity_u"]
        velo[:, where], velo_u[:, where] = subdata[:, where].layers["velocity"] + velo[:, where], \
                                           subdata[:, where].layers[
                                               "velocity_u"] + velo_u[:, where]
        adata.layers["velocity"], adata.layers["velocity_u"] = velo, velo_u


from typing import List

from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays

import numpy as np
from numpy import ndarray
from numpy.testing import assert_almost_equal, assert_array_equal

from scvelo.core import clipped_log, invert, prod_sum, sum


class TestClippedLog:
    @given(
        a=arrays(
            float,
            shape=st.integers(min_value=1, max_value=100),
            elements=st.floats(
                min_value=-1e3, max_value=1e3, allow_infinity=False, allow_nan=False
            ),
        ),
        bounds=st.lists(
            st.floats(
                min_value=0, max_value=100, allow_infinity=False, allow_nan=False
            ),
            min_size=2,
            max_size=2,
            unique=True,
        ),
        eps=st.floats(
            min_value=1e-6, max_value=1, allow_infinity=False, allow_nan=False
        ),
    )
    def test_flat_arrays(self, a: ndarray, bounds: List[float], eps: float):
        lb = min(bounds)
        ub = max(bounds) + 2 * eps

        a_logged = clipped_log(a, lb=lb, ub=ub, eps=eps)

        assert a_logged.shape == a.shape
        if (a <= lb).any():
            assert_almost_equal(np.abs(a_logged - np.log(lb + eps)).min(), 0)
        else:
            assert (a_logged >= np.log(lb + eps)).all()
        if (a >= ub).any():
            assert_almost_equal(np.abs(a_logged - np.log(ub - eps)).min(), 0)
        else:
            assert (a_logged <= np.log(ub - eps)).all()

    @given(
        a=arrays(
            float,
            shape=st.tuples(
                st.integers(min_value=1, max_value=100),
                st.integers(min_value=1, max_value=100),
            ),
            elements=st.floats(
                min_value=-1e3, max_value=1e3, allow_infinity=False, allow_nan=False
            ),
        ),
        bounds=st.lists(
            st.floats(
                min_value=0, max_value=100, allow_infinity=False, allow_nan=False
            ),
            min_size=2,
            max_size=2,
            unique=True,
        ),
        eps=st.floats(
            min_value=1e-6, max_value=1, allow_infinity=False, allow_nan=False
        ),
    )
    def test_2d_arrays(self, a: ndarray, bounds: List[float], eps: float):
        lb = min(bounds)
        ub = max(bounds) + 2 * eps

        a_logged = clipped_log(a, lb=lb, ub=ub, eps=eps)

        assert a_logged.shape == a.shape
        if (a <= lb).any():
            assert_almost_equal(np.abs(a_logged - np.log(lb + eps)).min(), 0)
        else:
            assert (a_logged >= np.log(lb + eps)).all()
        if (a >= ub).any():
            assert_almost_equal(np.abs(a_logged - np.log(ub - eps)).min(), 0)
        else:
            assert (a_logged <= np.log(ub - eps)).all()


class TestInvert:
    @given(
        a=arrays(
            float,
            shape=st.integers(min_value=1, max_value=100),
            elements=st.floats(max_value=1e3, allow_infinity=False, allow_nan=False),
        )
    )
    def test_flat_arrays(self, a: ndarray):
        a_inv = invert(a)

        if a[a != 0].size == 0:
            assert a_inv[a != 0].size == 0
        else:
            assert_array_equal(a_inv[a != 0], 1 / a[a != 0])

        if 0 in a:
            assert np.isnan(a_inv[a == 0]).all()
        else:
            assert set(a_inv[a == 0]) == set()

    @given(
        a=arrays(
            float,
            shape=st.tuples(
                st.integers(min_value=1, max_value=100),
                st.integers(min_value=1, max_value=100),
            ),
            elements=st.floats(max_value=1e3, allow_infinity=False, allow_nan=False),
        )
    )
    def test_2d_arrays(self, a: ndarray):
        a_inv = invert(a)

        if a[a != 0].size == 0:
            assert a_inv[a != 0].size == 0
        else:
            assert_array_equal(a_inv[a != 0], 1 / a[a != 0])

        if 0 in a:
            assert np.isnan(a_inv[a == 0]).all()
        else:
            assert set(a_inv[a == 0]) == set()


# TODO: Extend test to generate sparse inputs as well
# TODO: Make test to generate two different arrays a1, a2
# TODO: Check why tests fail with assert_almost_equal
class TestProdSum:
    @given(
        a=arrays(
            float,
            shape=st.integers(min_value=1, max_value=100),
            elements=st.floats(max_value=1e3, allow_infinity=False, allow_nan=False),
        ),
        axis=st.integers(min_value=0, max_value=1),
    )
    def test_flat_array(self, a: ndarray, axis: int):
        assert np.allclose((a * a).sum(axis=0), prod_sum(a, a, axis=axis))

    @given(
        a=arrays(
            float,
            shape=st.tuples(
                st.integers(min_value=1, max_value=100),
                st.integers(min_value=1, max_value=100),
            ),
            elements=st.floats(max_value=1e3, allow_infinity=False, allow_nan=False),
        ),
        axis=st.integers(min_value=0, max_value=1),
    )
    def test_2d_array(self, a: ndarray, axis: int):
        assert np.allclose((a * a).sum(axis=axis), prod_sum(a, a, axis=axis))


# TODO: Extend test to generate sparse inputs as well
class TestSum:
    @given(
        a=arrays(
            float,
            shape=st.integers(min_value=1, max_value=100),
            elements=st.floats(max_value=1e3, allow_infinity=False, allow_nan=False),
        ),
    )
    def test_flat_arrays(self, a: ndarray):
        a_summed = sum(a=a, axis=0)

        assert_array_equal(a_summed, a.sum(axis=0))

    @given(
        a=arrays(
            float,
            shape=st.tuples(
                st.integers(min_value=1, max_value=100),
                st.integers(min_value=1, max_value=100),
            ),
            elements=st.floats(max_value=1e3, allow_infinity=False, allow_nan=False),
        ),
        axis=st.integers(min_value=0, max_value=1),
    )
    def test_2d_arrays(self, a: ndarray, axis: int):
        a_summed = sum(a=a, axis=axis)

        if a.ndim == 1:
            axis = 0

        assert_array_equal(a_summed, a.sum(axis=axis))



from abc import ABC, abstractmethod
from typing import Dict, Tuple, Union

from numpy import ndarray


class DynamicsBase(ABC):
    @abstractmethod
    def get_solution(
        self, t: ndarray, stacked: True, with_keys: bool = False
    ) -> Union[Dict, Tuple[ndarray], ndarray]:
        """Calculate solution of dynamics.

        Arguments
        ---------
        t
            Time steps at which to evaluate solution.
        stacked
            Whether to stack states or return them individually. Defaults to `True`.
        with_keys
            Whether to return solution labelled by variables in form of a dictionary.
            Defaults to `False`.

        Returns
        -------
        Union[Dict, Tuple[ndarray], ndarray]
            Solution of system. If `with_keys=True`, the solution is returned in form of
            a dictionary with variables as keys. Otherwise, the solution is given as
            a `numpy.ndarray` of form `(n_steps, n_vars)`.
        """

        return

    @abstractmethod
    def get_steady_states(
        self, stacked: True, with_keys: False
    ) -> Union[Dict[str, ndarray], Tuple[ndarray], ndarray]:
        """Return steady state of system.

        Arguments
        ---------
        stacked
            Whether to stack states or return them individually. Defaults to `True`.
        with_keys
            Whether to return solution labelled by variables in form of a dictionary.
            Defaults to `False`.

        Returns
        -------
        Union[Dict[str, ndarray], Tuple[ndarray], ndarray]
            Steady state of system.
        """

        return
